# TODO mauybe hide green pointer if mouse is out of the image --> dangerous
# NB this will ultimately become the one and only paint widget in the app --> all others or most of them should disappear and be replaced by that
# TODO need add the save but only for the mask edit item!!!
import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
import traceback
from qtpy.QtGui import QKeySequence
from skimage.measure import label, regionprops
from qtpy.QtCore import QRect, QTimer, Qt
from qtpy.QtWidgets import QWidget, QApplication, QMessageBox
from epyseg.dialogs.opensave import saveFileDialog
from epyseg.draw.widgets.vectorial import VectorialDrawPane
from qtpy.QtWidgets import QMenu
from qtpy import QtCore, QtGui, QtWidgets
from epyseg.img import toQimage, Img, get_white_bounds, RGB_to_int24, is_binary, auto_scale, save_as_tiff
from epyseg.ta.measurements.measurements3D.get_point_on_surface_if_centroid_is_bad import point_on_surface
from epyseg.ta.segmentation.neo_wshed import wshed
from epyseg.ta.selections.selection import get_colors_drawn_over, convert_selection_color_to_coords
from epyseg.ta.tracking.tools import smart_name_parser
import numpy as np
from epyseg.tools.logger import TA_logger # logging

logger = TA_logger()

class Createpaintwidget(QWidget):
    # auto_convert_float_to_binary_threshold = 0.4

    default_mask_name = 'handCorrection.tif'

    # maybe store
    def __init__(self, enable_shortcuts=False): #, methods_overriding=None

        super().__init__()
        #
        # if methods_overriding is not None:
        #     for k,v in methods_overriding.items():
        #         # replace internal method by external one
        #         k = v
        # self.image_path = None
        self.save_path = None
        self.vdp = VectorialDrawPane(active=False) # the vectorial drawing panel
        self.raw_image = None # the original unprocessed image
        self.image = None # the displayed image (a qimage representation of raw_image can be a subset of raw_image_also)
        self.raw_mask = None  # the original unprocessed mask
        self.imageDraw = None # the mask image drawn over the image
        self.raw_user_drawing = None # contains just the user drawing without image mask
        self.cursor = None
        self.maskVisible = True
        self.scale = 1.0
        self.drawing = False
        self.brushSize = 3
        self.minimal_cell_size = 10
        self._clear_size = 30
        self.drawColor = QtGui.QColor(QtCore.Qt.red) # blue green cyan
        self.eraseColor = QtGui.QColor(QtCore.Qt.black)
        self.eraseColor_visualizer = QtGui.QColor(QtCore.Qt.green)
        self.cursorColor = QtGui.QColor(QtCore.Qt.green)
        self.lastPoint = QtCore.QPoint()
        self.change = False
        self.propagate_mouse_move = False
        # KEEP IMPORTANT required to track mouse even when not clicked
        self.setMouseTracking(True)  # KEEP IMPORTANT
        self.scrollArea = None
        self.statusBar = None
        self.drawing_enabled = True
        self.channel = None
        self.force_cursor_visible = False # required to force activate cursor to be always visible
        self.save_file_name=None
        self.multichannel_mode = False

        self.auto_convert_float_to_binary_threshold = 0.4

        # self.multi_channel_editor_n_save = False
        # TODO add sortcuts

        if enable_shortcuts:
            self.add_shortcuts() # TODO make this optional as it can cause trouble

    def force_cursor_to_be_visible(self, boolean):
        self.force_cursor_visible = boolean

        # can be used for saving
    def set_save_path(self, path):
        self.save_path = path

    def get_save_path(self):
        return self.save_path

    def set_scale(self, scale):
        # print(self.scale) # 0.10000000000000014 --> this is still ok --> this value crashes 1.3877787807814457e-16

        # make sure scale is bounded to avoid issues and non sense scalings
        if scale<0.01:
            scale = 0.01
        if scale>50:
            scale=50

        self.scale = scale
        self.vdp.scale =scale
        # DO I NEED ADJUST MORE SCALES ???
        # self.update()

    def get_scale(self):
        return self.scale

    # comprend pas ça marche pas si parent
    def add_shortcuts(self):
        # zoomPlus = QtWidgets.QShortcut("Ctrl+Shift+=", self)
        # zoomPlus.activated.connect(self.zoomIn)
        # zoomPlus.setContext(QtCore.Qt.ApplicationShortcut)  # make sure the shorctut always remain active
        #
        # zoomPlus2 = QtWidgets.QShortcut("Ctrl++", self)
        # zoomPlus2.activated.connect(self.zoomIn)
        # zoomPlus2.setContext(QtCore.Qt.ApplicationShortcut)  # make sure the shorctut always remain active
        #
        # zoomMinus = QtWidgets.QShortcut("Ctrl+Shift+-", self)
        # zoomMinus.activated.connect(self.zoomOut)
        # zoomMinus.setContext(QtCore.Qt.ApplicationShortcut)  # make sure the shorctut always remain active
        #
        # zoomMinus2 = QtWidgets.QShortcut("Ctrl+-", self)
        # zoomMinus2.activated.connect(self.zoomOut)
        # zoomMinus2.setContext(QtCore.Qt.ApplicationShortcut)

        # enter_pressed = QtWidgets.QShortcut("Ctrl+Shift+=", self)
        # zoomPlus.activated.connect(self.zoomIn)
        # zoomPlus.setContext(QtCore.Qt.ApplicationShortcut)  # make sure the shorctut always remain active


        padEnterShortcut = QtWidgets.QShortcut(QtGui.QKeySequence(QtCore.Qt.Key_Enter), self)  # voila
        padEnterShortcut.activated.connect(self.apply)
        padEnterShortcut.setContext(QtCore.Qt.ApplicationShortcut)  # make s

        enterShortcut = QtWidgets.QShortcut(QtGui.QKeySequence(QtCore.Qt.Key_Return), self)  # voila
        enterShortcut.activated.connect(self.apply)
        enterShortcut.setContext(QtCore.Qt.ApplicationShortcut)  # make s

        enterShortcut2 = QtWidgets.QShortcut('Shift+Return', self)  # voila
        enterShortcut2.activated.connect(self.apply)
        enterShortcut2.setContext(QtCore.Qt.ApplicationShortcut)  # make s

        # pb if parent has same shortcut then ignored
        # the pb is really with enter --> maybe stolen focus by dialog ????
        enterShortcut3 = QtWidgets.QShortcut('Shift+Enter', self)
        enterShortcut3.activated.connect(self.apply)
        enterShortcut3.setContext(QtCore.Qt.ApplicationShortcut)  # make sure the shorctut always remain active


        self.shrtM = QtWidgets.QShortcut("M", self)
        self.shrtM.activated.connect(self.m_apply)
        self.shrtM.setContext(QtCore.Qt.ApplicationShortcut)


        self.ctrl_shift_S_grab_screen_shot = QtWidgets.QShortcut('Ctrl+Shift+S', self)
        self.ctrl_shift_S_grab_screen_shot.activated.connect(self.grab_screen_shot)
        self.ctrl_shift_S_grab_screen_shot.setContext(QtCore.Qt.ApplicationShortcut)

        self.increase_contrastC = QtWidgets.QShortcut('C', self)
        self.increase_contrastC.activated.connect(self.increase_contrast)
        self.increase_contrastC.setContext(QtCore.Qt.ApplicationShortcut)

        # QtGui.QShortcut(
        # self.supr = QtWidgets.QShortcut(QtGui.QKeySequence(QtCore.Qt.Key_Delete), self)

        # QKeySequence(tr("Ctrl+X, Ctrl+C"))
        # QKeySequence(Qt.CTRL + Qt.Key_X, Qt.CTRL + Qt.Key_C)

        # self.supr = QtWidgets.QShortcut(QKeySequence(Qt.Key_Delete), self)
        # self.supr.activated.connect(self.suppr_pressed)
        # self.supr.setContext(QtCore.Qt.ApplicationShortcut)


        # connect the m shortcut to show and hide mask

    # this returns the coords of the cells drawn over --> very useful for clones, adding/removing cells such as dividing cells and dying cells !!!
    # that seems to work very well --> TODO finalize that now or soon
    def get_selection_coords_based_on_current_mask(self, bg_color=None):
        coords_of_selection = []

        # pb --> here I need do it on the labeled image -->

        img_to_analyze = self.get_raw_image()
        # si l'image est juste un mask binaire alors -->  faut la convertir en label --> TODO
        if is_binary(img_to_analyze):
            if len(img_to_analyze.shape)==3:
                # assume it's a handCorrection mask and the mask is 3channel
                img_to_analyze = img_to_analyze[...,0]
            img_to_analyze = label(img_to_analyze, connectivity=1, background=255) # will that always work
            bg_color = 0

        # why is this shit empty
        mask=self.get_user_drawing()

        # --> seems to be ok --> then why 0
        # import matplotlib.pyplot as plt
        # plt.imshow(img_to_analyze)
        # plt.show()

        selected_cells =  get_colors_drawn_over(mask,  img_to_analyze)
        if not selected_cells:
            logger.warning('No selection found --> nothing to do')
            return coords_of_selection

        self.set_mask(np.zeros_like(mask))
        self.update()

        tmp = img_to_analyze
        if len(tmp.shape) == 3:
            tmp = RGB_to_int24(tmp)
        #
        if bg_color is None:
            if 0xFFFFFF in tmp:
                bg_color = 0xFFFFFF
            else:
                bg_color = 0
        #
        # cell_label = label(tmp, connectivity=1, background=bg_color)
        #
        # # for all the selected cells --> return the coords --> probably a quite good idea...
        # for region in regionprops(cell_label):
        #     color = tmp[region.coords[0][0], region.coords[0][1]]
        #     if color in selected_cells:
        #         point_awlays_inside_the_cell = point_on_surface(region, cell_label)
        #         coords_of_selection.append(point_awlays_inside_the_cell)
        #
        # return coords_of_selection

        return convert_selection_color_to_coords(tmp, selected_cells=selected_cells, bg_color=bg_color)

    # added this so that I can easily get user selection using a cell ID
    def get_colors_drawn_over(self, forbidden_colors=[0,0xFFFFFF]):
        mask = self.get_user_drawing()
        # print('I was called')
        if mask is None:
            return

        selected_colors = get_colors_drawn_over(mask,  self.get_raw_image(),forbidden_colors=forbidden_colors)
        # # print(mask)
        # # need get all the colors of the raw image that correspond to user selection
        # selected_colors = self.get_raw_image()[mask != 0]
        # selected_colors = RGB_to_int24(selected_colors)
        # selected_colors = set(selected_colors.ravel().tolist())
        # # if 0xFFFFFF in selected_colors:
        # #     selected_colors.remove(0xFFFFFF)
        # if forbidden_colors is not None:
        #     if not isinstance(forbidden_colors, list):
        #         forbidden_colors = [forbidden_colors]
        #     for color in forbidden_colors:
        #         if color in selected_colors:
        #             selected_colors.remove(color)
        # selected_colors = list(selected_colors)
        self.set_mask(np.zeros_like(mask))
        self.update()
        return selected_colors

    # def suppr_pressed(self):

    def increase_contrast(self):
        # print('increasing contrast')
        try:
            # metadata = self.raw_image.metadata
            # sqqsdq
            # self.raw_image = Img(auto_threshold(self.raw_image),metadata=metadata)
            # self.set_image(self.raw_image)
            # shall I do it for the displayed image only ???? --> probably smarter porbably shoudl do a get display image that is independent of all other functions and handles all dims
            # very dumb --> needs be done just for the dipslayed image and nothing else and --> maybe get the image back from the screen directly --> would be smarted and better

            # maybe do this only on the displayed image --> using convert_qimage_to_numpy or do something that gets the image to be displayed including the channels
            # or do a tool that will recover the image of interest with channels and also the rest of the dimensions --> MEGA TODO !!!
            print('auto-increase contrast')
            # hack to preserve luts --> TODO
            copy = auto_scale(np.copy(self.raw_image))
            meta = None
            try:
                meta=self.raw_image.metadata
            except:
                pass
            # print(self.raw_image.metadata)
            self.set_display(copy, metadata=meta)
        except:
            traceback.print_exc()

    def grab_screen_shot(self):
            # screenshot = QPixmap.grabWindow(self.winId())
            output_file = saveFileDialog(parent_window=self, extensions="Supported Files (*.png);;All Files (*)",
                                         default_ext='.png')
            if output_file is not None:

                try:
                    # we hide the cursor
                    self.cursor = QtGui.QImage(self.image.size(), QtGui.QImage.Format_ARGB32)
                    self.cursor.fill(QtCore.Qt.transparent)
                    self.update()
                    screenshot=self.grab()
                    screenshot.save(output_file, 'png')
                    # restore cursor
                    # or paint without cursor --> TODO do that better some day
                except:
                    logger.error('Could not grab a screenshot...')

    def save(self):
        # save_img
        print('ctrl S save method to override')

    def apply(self):
        print('apply method to override')

    def shift_apply(self):
        print('shift apply method to override')

    def ctrl_m_apply(self):
        print('ctrl m apply method to override')

    def suppr_pressed(self):
        # print('Suppr method to override')
        if self.vdp.active:
            self.vdp.removeCurShape()
            self.update()
        else:
            print('Suppr method to override')


    # def m_apply(self):
    #     print('m apply method to override')
    #     qsdsqqsdsq
    def m_apply(self):
        # print('m apply called')
        self.maskVisible = not self.maskVisible
        self.update()

    def set_display(self, display, metadata=None):
        # this actually sets the displayed image --> different from set imlage, can be used to show channels instead or stuff like that instead of the actula imge
        if isinstance(display, np.ndarray):
            display = toQimage(display, metadata=metadata)
        self.image = display
        self.update()

    def set_image(self, img):
        if img is None:
            self.save_file_name = None
            self.raw_image = None
            self.image = None
            self.imageDraw = None
            self.raw_mask = None
            self.raw_user_drawing = None
            self.unsetCursor()  # restore cursor
            self.update()
            return
        else:
                # remove cursor only if image is shown and not vectorial drawing mode!
                if self.vdp.active:
                    self.unsetCursor()
                else:
                    self.setCursor(Qt.BlankCursor)
                # try:
                #     if isinstance(img, np.ndarray):
                #         self.image = img.getQimage() # bug is here
                #         # set an empty mask
                #         self.imageDraw = QtGui.QImage(self.image.size(), QtGui.QImage.Format_ARGB32)
                #         self.imageDraw.fill(QtCore.Qt.transparent)
                #     else:
                #         self.image = img[0].getQimage()
                #         # set the mask
                #         self.imageDraw = Img(self.createRGBA(img[1].getQimage()), dimensions='hwc').getQimage()
                # except:
                # NB THIS IS VERY REDUNDNAT ---> STICK TO ONE METHOD
                if isinstance(img, str):
                    # self.image_path = img# maybe this image path can be used for saving, could also be provided serparately
                    self.save_file_name = smart_name_parser(img, ordered_output=self.default_mask_name)
                    img = Img(img)

                # self.image_path=None # maybe this image path can be used for saving
                if isinstance(img, np.ndarray):

                    # bug is exactly here --> why and where

                    # dimensions = None
                    # if isinstance(img, Img):
                    #     dimensions = img.get_dimensions_as_string()

                    # self.image_path = None
                    self.raw_image = img

                    # print('tests',img.shape, isinstance(img,Img))

                    self.image = toQimage(img)# maybe this image path can be used for saving
                    if self.image is None:
                        logger.error('Image could not be displayed...')
                        self.unsetCursor()
                        self.update()
                        return
                    self.imageDraw = QtGui.QImage(self.image.size(), QtGui.QImage.Format_ARGB32)
                    self.imageDraw.fill(QtCore.Qt.transparent)
                    self.raw_user_drawing =QtGui.QImage(self.image.size(), QtGui.QImage.Format_ARGB32)
                    self.raw_user_drawing.fill(QtCore.Qt.transparent)
                else:
                    # self.image_path = None
                    self.image = toQimage(img[0])
                    self.raw_image = img[0]  #
                    # same as set mask --> how to do that
                    self.imageDraw = toQimage(Img(self.createRGBA(img[1]), dimensions='hwc'),preserve_alpha=True)#.getQimage()
                    self.raw_user_drawing = QtGui.QImage(self.imageDraw.size(), QtGui.QImage.Format_ARGB32)
                    self.raw_user_drawing.fill(QtCore.Qt.transparent)

        if self.force_cursor_visible:
            self.unsetCursor()

        # self.image = QPixmap(100,200).toImage()
        width = self.image.size().width()
        height = self.image.size().height()
        top = self.geometry().x()
        left = self.geometry().y()
        self.setGeometry(top, left, width*self.scale, height*self.scale)

        self.cursor = QtGui.QImage(self.image.size(), QtGui.QImage.Format_ARGB32)
        self.cursor.fill(QtCore.Qt.transparent)
        self.update()

    def get_raw_image(self):
        return self.raw_image

    def binarize(self, mask, auto_convert_float_to_binary=None, force=False):
        if auto_convert_float_to_binary is None:
            auto_convert_float_to_binary = self.auto_convert_float_to_binary_threshold
        if auto_convert_float_to_binary and (mask.max() <= 1 or force):
            # print('autoconvert')
            mask = mask > auto_convert_float_to_binary
        return mask

    # do I need that ??? maybe yes if I just wanna change the mask
    def set_mask(self, mask, auto_convert_float_to_binary=None):
        if auto_convert_float_to_binary is None:
            auto_convert_float_to_binary = self.auto_convert_float_to_binary_threshold
        if isinstance(mask, str):
            self.save_file_name = mask
            # self.image_path = img# maybe this image path can be used for saving, could also be provided serparately
            mask = Img(mask)
            self.raw_mask = mask
            # for TA compat
            if mask.has_c():
                # but should also save the raw mask as for the raw_image --> TODO --> maybe
                if self.channel is not None:
                    mask=mask[...,self.channel]
                else:
                    mask = mask[..., 0]

        # convert mask to image draw
        if mask is None:
            self.imageDraw = None
            self.raw_user_drawing = None
            self.raw_mask = None
        else:
            # auto convert float deep learning masks to binary --> maybe this should be a parameter ...
            # print('tada', auto_convert_float_to_binary, mask.max())

            mask = self.binarize(mask, auto_convert_float_to_binary=auto_convert_float_to_binary)


            # print('mask.shape',mask.shape)
            # convert mask to a mask
            # self.imageDraw = QtGui.QImage(self.image.size(), QtGui.QImage.Format_ARGB32)
            # self.imageDraw.fill(QtCore.Qt.transparent)

            # print('in here tmp ', mask.shape)


            # if image has z channels and is more than just 2D / is 3D I could make this smarter by loading just the right image --> TODO --> sync the channel here with the channel of the currently dispalyer image
            # SHALL I MAKE A FULL GUI FOR THAT ON RELY ON WHAT I HAVE ALREADY !!!!
            self.imageDraw = toQimage(Img(self.createRGBA(mask), dimensions='hwc'),preserve_alpha=True) #.getQimage()  # marche pas car besoin d'une ARGB
            self.raw_user_drawing = QtGui.QImage(self.imageDraw.size(), QtGui.QImage.Format_ARGB32)
            self.raw_user_drawing.fill(QtCore.Qt.transparent) # somehow this is really required to have an empty image otherwise it is not --> ??? why
        self.update()


    # do I have a way to just get what was drawn by the user and the bounds of it --> maybe by substratction --> or can I get the difference
    # I need bounds of drawn stuff --> think about how to do that !!!
    # we have used the user drawing for something --> let's reset it
    # shall I also save the raw mask unperturbed by user drawing --> can be useful too in fact to do things with it
    def reset_user_drawing(self):
        if self.imageDraw is None:
            self.raw_user_drawing = None
        else:
            self.raw_user_drawing = QtGui.QImage(self.imageDraw.size(), QtGui.QImage.Format_ARGB32)
            self.raw_user_drawing.fill(QtCore.Qt.transparent)

            # shall I draw any user interaction also in a separate file so that I can get just what was drawn without having the original mask ??? --> maybe a good idea --> call it get user drawing ???
    def get_mask(self):
        # nb will only work for red pen !!!! --> do not allow to change pen color
        # pas mal mais image est RGBA
        if self.imageDraw is not None:
            # return Img.qimageToNumpy(self.imageDraw, mode='bnw8')
            mask = self.convert_qimage_to_numpy(self.imageDraw)[..., 2]
            # for some reason there are some 1 in the image --> needs a fix
            mask[mask!=255]=0
            return mask
        else:
            return None

    def get_user_drawing(self, show_erased=False):
        if self.raw_user_drawing is not None:



            # bug here the user draing contains the original mask --> bug

            all = self.convert_qimage_to_numpy(self.raw_user_drawing)

            # plt.imshow(all)
            # plt.show()

            # no clue why this stuff is empty
            # import matplotlib.pyplot as plt
            # plt.imshow(all)
            # plt.show()

            drawn = all[..., 2]


            # plt.imshow(erased) #--> ok
            # plt.show()

            # plt.imshow(drawn)
            # plt.show()


            drawn[drawn != 255] = 0

            if show_erased:
                erased = all[..., 1]
                erased[erased != 255] = 0
                return [erased, drawn]

            return drawn
        else:
            return None

    # from https://stackoverflow.com/questions/19902183/qimage-to-numpy-array-using-pyside
    # nb there seems to be a bug in the order of the channels
    def convert_qimage_to_numpy(self, qimage):
        qimage = qimage.convertToFormat(QtGui.QImage.Format.Format_RGB32)
        width = qimage.width()
        height = qimage.height()
        image_pointer = qimage.bits() # creates a deep copy --> this is what I want
        try:
            image_pointer.setsize(qimage.sizeInBytes()) # qt6 version of the stuff
        except:
            image_pointer.setsize(qimage.byteCount())


        # arr = np.array(image_pointer,copy=True).reshape(height, width, 4)
        arr = np.array(image_pointer).reshape(height, width, 4)
        return arr

    def createRGBA(self, handCorrection):





        # print('Im called')
        # use pen color to display the mask
        # in fact I need to put the real color
        # nb this will definitely not work if the user wants to have full 3D stuff

        # shall I also offer the zooming on the objects --> ????




        RGBA = np.zeros((handCorrection.shape[0], handCorrection.shape[1], 4), dtype=np.uint8)

        red = self.drawColor.red()
        green = self.drawColor.green()
        blue = self.drawColor.blue()

        #(handCorrection.shape[0], handCorrection.shape[1], 4))
        # mask = handCorrection[handCorrection==255]

        # ce truc est bleu --> pkoi
        # RGBA[..., 0] = 255#handCorrection --> totalement transp --> transparence color --> GREEN --> BLUE
        # RGBA[..., 1] = 0  # this is the red channel --> RED --> GREEN
        # RGBA[..., 2] = 0 # rien ici --> tt transparent --> alpha ??? --> ALPHA -->RED
        # RGBA[..., 3] = 255 # --> BLUE --> ALPHA --> ok

        # pkoi je swappe les channels avant

        # 255 partout --> blanc
        # 0 255 255 0 --> rouge --> 1 = red channel and 2 = alpha --> weird

        # RGBA[RGBA[..., 0] > 0, 0] = blue
        # RGBA[RGBA[..., 1] > 0, 1] = green
        # RGBA[RGBA[..., 2] > 0, 2] = red
        # RGBA[..., 3] = 255 # ne s'affiche pas et je comprend pas pkoi

        #BGR in fact --> need fix here or the other --> the one in


        # how can I do that in a simpler way ???
        # bug somewhere in qimage --> fix it some day --> due to bgra instead of RGBA
        # assumes image has just the first
        RGBA[handCorrection != 0, 0] = blue # b
        # RGBA[..., 1] = handCorrection
        RGBA[handCorrection != 0, 1] = green # g
        RGBA[handCorrection != 0, 2] = red # r
        RGBA[..., 3] = 255 # alpha --> indeed alpha
        RGBA[handCorrection == 0, 3] = 0 # ça marche maintenant --> super complexe qd meme je trouve

        # can I do it in one go --> I guess yes!!!
        # RGBA[..., 3] = 255
        # RGBA[handCorrection != 0] = blue,green,red, 0
        # RGBA[handCorrection==0, 3] = 255

        # ok mais plus transparent

        #
        # print(red, green, blue)

        # RGBA[RGBA[..., 0] == 255] = red
        # RGBA[RGBA[..., 1] == 255] = green
        # RGBA[RGBA[..., 2] == 255] = blue

        return RGBA

        # TODO also implement channel change directly within the display tool
        # if merge is applied --> apply on average of all channels --> maybe not so smart an idea but ok to start with and better than what I do in TA

    # dunno if I put it here or not ???
    # def get_nb_channels(self):
    #     if self.

    # depreacted --> ideally should not be used
    # TODO--> maybe implement a multi channel save -> TODO
    # skip
    def channelChange(self, i, skip_update_display=False):
        # update displayed image depending on channel
        # dqqsdqsdqsd
        # pass
        # try change channel if

        # print('in channel change !!!')
        # if image_to_display is None:
        #     image_to_display = self.raw_image
        if self.raw_image is not None:
            # print('in', self.img.metadata)
            # if self.Stack.currentIndex() == 0:
                # need copy the image --> implement that
                # print(self.img[..., i].copy())
                # print(self.img[..., i])
                if i == 0 and not skip_update_display:
                    self.set_display(self.raw_image)
                    self.channel = None
                    # print('original', self.img.metadata)
                else:
                    if not skip_update_display:
                        # print('modified0', self.img.metadata)
                        # I need a hack when the image is single channel yet I need several masks for it !!!
                        meta = None
                        try:
                            meta = self.raw_image
                        except:
                            pass
                        if self.multichannel_mode and i-1>=self.raw_image.shape[-1]:
                            channel_img = self.raw_image.imCopy(c=0) # if out of bonds load the first channel
                        else:
                            channel_img = self.raw_image.imCopy(c=i - 1)  # it's here that it is affected
                        self.set_display(
                                channel_img, metadata=meta)  # maybe do a set display instead rather --> easier to handle --> does a subest of the other
                    self.channel = i-1
                    # print('modified1', self.img.metadata)
                    # print('modified2', channel_img.metadata)

                    if self.multichannel_mode:
                        if self.raw_mask is not None:
                            # print('multichannel_mode', self.multichannel_mode)
                            try:
                                # print('self.raw_mask.shape',self.raw_mask.shape)
                                if len(self.raw_mask.shape)<=2:
                                    self.raw_mask = self.raw_mask[..., np.newaxis]
                                self.set_mask(self.raw_mask[..., i-1])
                            except:
                                # print(i-1,  self.raw_mask.shape[-1] , len(self.raw_mask.shape))
                                if i-1 >= self.raw_mask.shape[-1] and len(self.raw_mask.shape)>2:
                                    # extend mask


                                    if len(self.raw_mask.shape) > 2:
                                        raw_mask2 = np.zeros_like(self.raw_mask, shape=(*self.raw_mask.shape[0:-1], i))
                                    else:
                                        raw_mask2 = np.zeros_like(self.raw_mask, shape=(*self.raw_mask.shape, i))

                                    # copy existing mask to it
                                    # print(raw_mask2.shape, self.raw_mask.shape)
                                    for ch in range(self.raw_mask.shape[-1]):
                                        raw_mask2[...,ch]=self.raw_mask[...,ch]
                                    self.raw_mask = raw_mask2
                                    self.set_mask(self.raw_mask[..., i - 1])
                                else:
                                    traceback.print_exc()
                        else:
                            # create empty mask
                            if len(self.raw_image.shape)>2:
                                self.raw_mask = np.zeros(shape=(*self.raw_image.shape[0:-1],i))
                            else:
                                self.raw_mask = np.zeros(shape=(*self.raw_image.shape, i))

                            # print('doubs', self.raw_mask.shape, self.raw_image.shape)
                            self.set_mask(self.raw_mask[..., i - 1])
                self.update()
        else:
            self.channel = None
            # else:
            #     # logger.error("Not implemented yet TODO add support for channels in 3D viewer")
            #     # sdqdqsdsqdqsd
            #     self.loadVolume()


    def get_colors_drawn_over(self):
        mask = self.get_user_drawing()
        # print('I was called')
        if mask is None:
            return []
        # print(mask)

        # need get all the colors of the raw image that correspond to user selection
        selected_colors =self.get_raw_image()[mask!=0]
        selected_colors = RGB_to_int24(selected_colors)

        selected_colors =set(selected_colors.ravel().tolist())
        if 0xFFFFFF in selected_colors:
            selected_colors.remove(0xFFFFFF)
        selected_colors = list(selected_colors)
        # this is the only way I found to reset the user drawing but this is dirty and I must find a better  way
        self.set_mask(np.zeros_like(mask))

        self.update()
        return selected_colors

    def edit_drawing(self):
        # if self.multichannel_mode and self.channel is None:
        #     logger.error('please select a channel first')
        #     return

        # basic mask drawing edit
        drawn_mask = self.get_mask()
        if drawn_mask is None:
            return
        drawn_mask = np.copy(drawn_mask)
        erased, drawn = self.get_user_drawing(show_erased=True)
        # bounds = get_white_bounds([drawn, erased])
        # if bounds is None:
        # pass
        # print(drawn_mask.shape)
        # print(drawn_mask.max(), drawn_mask.min())
        # print(drawn_mask.dtype)

        drawn_mask[erased!=0]=0
        drawn_mask[drawn!=0]=drawn[drawn!=0]
        if self.multichannel_mode:
            if self.raw_mask is not None:
                if self.channel is not None:
                    # print('overwriting mask in image')
                    self.raw_mask[..., self.channel]=drawn_mask
                    # big bug here this stuff contains the bg image --> really huge bug
                    # Img(self.raw_mask, dimensions='hwc').save('/E/Sample_images/sample_images_PA/test_complete_wing_raphael/tst2.tif')
                else:
                    # if nothing is specified assume ch 0
                    if len(self.raw_mask.shape)>2:
                        self.raw_mask[..., 0] = drawn_mask
                    else:
                        self.raw_mask = drawn_mask

        # Img(drawn_mask, dimensions='hw').save('/E/Sample_images/sample_images_PA/test_complete_wing_raphael/tst.tif')
        self.set_mask(Img(drawn_mask, dimensions='hw'))


    # draws/removes bonds locally à la TA
    def apply_drawing(self, minimal_cell_size=0):
        # print('minimal_cell_size', minimal_cell_size)

        drawn_mask = self.get_mask()
        # print(drawn_mask)
        # no image or nothing drawn --> return
        if drawn_mask is None:
            return
        erased, drawn = self.get_user_drawing(show_erased=True)
        bounds = get_white_bounds([drawn, erased])
        if bounds is None:
            # in case the guy presses shift apply with no mask then run it over the whole image --> can make sense to have this
            if minimal_cell_size is not None and minimal_cell_size !=0:
                # dirty way of changing size --> needs be improved
                minimal_cell_size = self.minimal_cell_size


                # print('minimal_cell_size2', minimal_cell_size)
                drawn_mask = wshed(drawn_mask, seeds='mask', min_seed_area=minimal_cell_size)
                self.set_mask(drawn_mask)
            return

        if np.count_nonzero(erased) != 0:
            lab_erased = label(erased, connectivity=2, background=0)
            rps_lab_erased = regionprops(lab_erased)

            lab_drawn_mask = label(drawn_mask, connectivity=1, background=255)
            rps_lab_drawn_mask = regionprops(lab_drawn_mask)

            for erased_bond in rps_lab_erased:
                ids = np.unique(lab_drawn_mask[erased_bond.coords[:, 0], erased_bond.coords[:, 1]])
                for id in ids:
                    if id != 0:
                        cur_bounds = rps_lab_drawn_mask[id - 1].bbox
                        bounds[0] = min(bounds[0], cur_bounds[0])
                        bounds[1] = max(bounds[1], cur_bounds[2])
                        bounds[2] = min(bounds[2], cur_bounds[1])
                        bounds[3] = max(bounds[3], cur_bounds[3])

        min_y, max_y, min_x, max_x = bounds

        first_x = 128
        first_y = 128
        min_x = min_x - 128
        max_x = max_x + 128
        min_y = min_y - 128
        max_y = max_y + 128

        # TODO fix upper bond too

        if min_x < 0:
            first_x = 128 + min_x
            min_x = 0
        if min_y < 0:
            first_y = 128 + min_y
            min_y = 0
        if max_y > drawn_mask.shape[0]:
            max_y = drawn_mask.shape[0]
        if max_x > drawn_mask.shape[1]:
            max_x = drawn_mask.shape[1]

        minished = wshed(drawn_mask[min_y:max_y, min_x:max_x], seeds='mask', min_seed_area=minimal_cell_size)
        drawn_mask[min_y:max_y, min_x:max_x][first_y:min(first_y + (bounds[1] - bounds[0]) + 1, minished.shape[0]),first_x:min(first_x + (bounds[3] - bounds[2]) + 1, minished.shape[1])] = minished[first_y:min(first_y + (bounds[1] - bounds[0]) + 1, minished.shape[0]), first_x:min(first_x + (bounds[3] - bounds[2]) + 1, minished.shape[1])]
        self.set_mask(drawn_mask)

    def force_consistent_range(self, img, auto_convert_float_to_binary=None):
        if auto_convert_float_to_binary is None:
            auto_convert_float_to_binary = self.auto_convert_float_to_binary_threshold
        # for ch in range(img.shape[-1]):
        #     tmp = img[...,ch]
        #     min = tmp.min()
        #     max = tmp.max()
        #     # if min!=max:
        #         # change range of the stuff
        #         # print('error in range')
        #     if max<=1.0:
        #         print('error in range')
        #     print(min, max)
        #     print(min, max)
        # img = 
        return self.binarize(img, auto_convert_float_to_binary=auto_convert_float_to_binary, force=True)

    def save_mask(self, multichannel_save=False, forced_nb_of_channels=None):
        # print('saving to ....' + self.save_file_name)
        if self.save_file_name is not None:
            # print('really saving/...')
            mask = self.get_mask()
            if mask is not None:
                logger.debug('saving mask to '+str(self.save_file_name))
                try:
                    single_channel_image = len(self.raw_image.shape)==2
                except:
                    single_channel_image = False
                if multichannel_save == False or single_channel_image  and (forced_nb_of_channels is None or forced_nb_of_channels ==0):
                    Img(mask, dimensions='hw').save(self.save_file_name)
                else:
                    if self.raw_mask is not None:
                        # print('quick saving edited mask')
                        # Img(self.raw_mask, dimensions='hwc').save(self.save_file_name)

                        if len(self.raw_mask.shape)>2:
                            # I need check all channels have the same normalization
                            Img(self.force_consistent_range(self.raw_mask).astype(np.uint8)*255, dimensions='hwc').save(self.save_file_name)
                        else:
                            Img(self.raw_mask, dimensions='hw').save(self.save_file_name)
                        return

                    # print('test', self.channel )
                    # print('orig img channels', self.raw_image.shape[-1])
                    # if self.channel is None:
                    #     logger.error('Please select a channel first')
                    #     return
                    # BELOW IS CRAPPY CODE --> BEST is if multichannel_mode to create a black empty image
                    # if os.path.exists(self.save_file_name):
                    #     try:
                    #         mask_out = Img(self.save_file_name)
                    #         if self.channel is not None:
                    #             mask_out[...,self.channel]=mask
                    #         else:
                    #             # assume ch0
                    #             mask_out[..., 0] = mask
                    #         Img(mask_out, dimensions='hwc').save(self.save_file_name)
                    #         return
                    #     except:
                    #         traceback.print_exc()
                    #     print('le fichier existe')
                    #
                    #     # on le charge
                    # # else:
                    # print('le fichier existe pas')
                    # mask_out = np.zeros_like(mask, shape=(*mask.shape, self.raw_image.shape[-1]))
                    # if self.channel is not None:
                    #     mask_out[..., self.channel] = mask
                    # else:
                    #     # assume ch0
                    #     mask_out[..., 0] = mask
                    # Img(mask_out, dimensions='hwc').save(self.save_file_name)
                    # sinon on recup le nb de canaux de l'image finale


    # def _update_channel(self, output, mask, channel):
    #     if channel is not None:
    #         mask_out = np.zeros_like(mask, shape=(*mask.shape, self.raw_image.shape[-1]))
    #         mask_out[..., self.channel] = mask
    #         Img(mask_out, dimensions='hwc').save(self.save_file_name)





    # this is the local seeded watershed à la TA
    def manually_reseeded_wshed(self):

        if self.get_raw_image() is None:
            return

        try:
            if self.channel is None and self.get_raw_image().has_c():
                logger.error('Please select a channel fisrt')
                return
        except:
            print(self.get_raw_image().shape, self.channel)
            if self.channel is None and len(self.get_raw_image().shape)>=3 and self.get_raw_image().shape[-1]>1:
                logger.error('Please select a channel fisrt')
                return

        usr_drawing = self.get_user_drawing()
        labs = label(usr_drawing, connectivity=2, background=0)

        # pb here because the user drawing also contains the mask and I don't want it
        # plt.imshow(labs)
        # plt.show()

        rps_user_drawing = regionprops(labs)
        if len(rps_user_drawing) == 0:
            return

        drawn_mask = self.get_mask()

        # plt.imshow(drawn_mask)
        # plt.show()

        drawn_mask[labs != 0] = 0
        lab_cells_user_drawing = label(drawn_mask, connectivity=1, background=255)
        rps_cells_user_drawing = regionprops(lab_cells_user_drawing)

        min_y = 100000000
        min_x = 100000000
        max_y = 0
        max_x = 0
        for rps_user in rps_user_drawing:
            label_id = lab_cells_user_drawing[rps_user.coords[0][0], rps_user.coords[0][1]]
            bbox = rps_cells_user_drawing[label_id - 1].bbox
            min_y = min(bbox[0], min_y)
            min_x = min(bbox[1], min_x)
            max_y = max(bbox[2], max_y)
            max_x = max(bbox[3], max_x)




        # here I need get the channel in fact
        img = self.get_raw_image()[min_y:max_y + 1, min_x:max_x + 1]
        if self.channel is not None:
            img = img[..., self.channel]
        minished = wshed(img,
                         seeds=labs[min_y:max_y + 1, min_x:max_x + 1], weak_blur=1.)
        drawn_mask[min_y:max_y + 1, min_x:max_x + 1][minished != 0] = minished[minished != 0]
        # plt.imshow(drawn_mask)
        # plt.show()
        self.set_mask(drawn_mask)


    def mousePressEvent(self, event):
        if not self.hasMouseTracking() or not self.drawing_enabled:
            return
        self.clickCount = 1
        if self.vdp.active:
            self.vdp.mousePressEvent(event)
            self.update()
            return

        if event.buttons() == QtCore.Qt.LeftButton or event.buttons() == QtCore.Qt.RightButton:
            self.drawing = True
            zoom_corrected_pos = event.pos() / self.scale
            self.lastPoint = zoom_corrected_pos
            self.drawOnImage(event)
        else:
            self.drawing = False


    # can I pass this event to something else to synchronize the mouse moves
    def mouseMoveEvent(self, event):
        if not self.hasMouseTracking() or not self.drawing_enabled:
            return
        else:
            if self.propagate_mouse_move == True:
                self.propagate_mouse_event(event)

        # print('in mouse move', self.hasMouseTracking(), self.drawing, self.vdp.active)
        if self.statusBar:
            zoom_corrected_pos = event.pos() / self.scale
            self.statusBar.showMessage('x=' + str(zoom_corrected_pos.x()) + ' y=' + str(
                zoom_corrected_pos.y()))
        if self.vdp.active:
            self.vdp.mouseMoveEvent(event)
            region = self.scrollArea.widget().visibleRegion()
            self.update(region)
            return
        self.drawOnImage(event)

    def mouseReleaseEvent(self, event):
        self.drawing = False
        if not self.hasMouseTracking() or not self.drawing_enabled:
            return
        if self.vdp.active:
            self.vdp.mouseReleaseEvent(event)
            self.update()  # required to update drawing
            return
        # if event.button == QtCore.Qt.LeftButton:
        #     self.drawing = False
        if self.clickCount == 1:
            QTimer.singleShot(QApplication.instance().doubleClickInterval(),
                              self.updateButtonCount)

    def propagate_mouse_event(self, event):
        print(self,event)
        pass

    def drawOnImage(self, event):
        # bug fix for image drawing when image is none
        if self.imageDraw is None:
            return
        zoom_corrected_pos = event.pos() / self.scale
        if self.drawing and (event.buttons() == QtCore.Qt.LeftButton or event.buttons() == QtCore.Qt.RightButton):
            # now drawing or erasing over the image
            self._draw_on_image(self.imageDraw, event, zoom_corrected_pos)
            # this contains only the user drawing and nothing else --> can be useful to have too!!!
            self._draw_on_image(self.raw_user_drawing, event, zoom_corrected_pos, erase_color=self.eraseColor_visualizer)



        # Drawing the cursor TODO add boolean to ask if drawing cursor should be shown
        painter = QtGui.QPainter(self.cursor)
        try:
            # We erase previous pointer
            r = QtCore.QRect(QtCore.QPoint(), self._clear_size * QtCore.QSize() * self.brushSize)
            painter.save()
            r.moveCenter(self.lastPoint)
            painter.setCompositionMode(QtGui.QPainter.CompositionMode_Clear)
            painter.eraseRect(r)
            painter.restore()
            # draw the new one

            stroke_size = 2
            if self.brushSize<6:
                stroke_size = 1

            # if is 1 --> I should really draw a point instead
            # I should also hide the cursor --> that would help
            painter.setPen(QtGui.QPen(self.cursorColor, stroke_size, QtCore.Qt.SolidLine, QtCore.Qt.RoundCap,
                                      QtCore.Qt.RoundJoin))
            if self.brushSize >1:
                painter.drawEllipse(zoom_corrected_pos, self.brushSize / 2.,
                                self.brushSize / 2.)
            else:
                painter.drawPoint(zoom_corrected_pos)
        except:
            traceback.print_exc()
        finally:
            painter.end()
        try:
            region = self.scrollArea.widget().visibleRegion()
        except:
            # assume no scroll area --> use self visible region
            region =self.visibleRegion()

        self.update(region)

        # required to erase mouse pointer
        self.lastPoint = zoom_corrected_pos


    def _draw_on_image(self, image_to_be_drawn, event, zoom_corrected_pos, erase_color=None):
        painter = QtGui.QPainter(image_to_be_drawn)
        try:
            # small hack to allow use pen or fingers on touch screens to erase when Ctrl or shift is pressed along with the click --> MAYBE PUT THIS AS AN OPTION SOME DAY BECAUSE THIS MAY PERTURB PEOPLE
            if event.buttons() == QtCore.Qt.LeftButton and not (event.modifiers() == QtCore.Qt.ControlModifier or event.modifiers() ==QtCore.Qt.ShiftModifier):
                painter.setPen(QtGui.QPen(self.drawColor, self.brushSize, QtCore.Qt.SolidLine, QtCore.Qt.RoundCap,
                                          QtCore.Qt.RoundJoin))
            else:
                if erase_color is None:
                    erase_color = self.eraseColor
                painter.setPen(QtGui.QPen(erase_color, self.brushSize, QtCore.Qt.SolidLine, QtCore.Qt.RoundCap,
                                          QtCore.Qt.RoundJoin))
            if self.lastPoint != zoom_corrected_pos:
                painter.drawLine(self.lastPoint, zoom_corrected_pos)
            else:
                # if zero length line then draw point instead
                painter.drawPoint(zoom_corrected_pos)
        except:
            traceback.print_exc()
        finally:
            painter.end()

    # adds context/right click menu but only in vectorial mode
    def contextMenuEvent(self, event):
        if not self.vdp.active:
            return
        cmenu = QMenu(self)
        newAct = cmenu.addAction("New")
        opnAct = cmenu.addAction("Open")
        quitAct = cmenu.addAction("Quit")
        action = cmenu.exec_(self.mapToGlobal(event.pos()))
        if action == quitAct:
            sys.exit(0)

    def updateButtonCount(self):
        self.clickCount = 1

    def mouseDoubleClickEvent(self, event):
        self.clickCount = 2

        # TODO maybe add controls to be sure we are in drawing mode with a pen ???
        # double click fill or delete
        if not self.vdp.active:
            zoom_corrected_pos = event.pos() / self.scale
            if event.buttons() == QtCore.Qt.LeftButton:
                # double click fill
                if self.dialog_fill_erase('fill'):
                    self.double_click_fill(zoom_corrected_pos)
            elif event.buttons() == QtCore.Qt.RightButton:
                # double click delete shape
                if self.dialog_fill_erase('erase'):
                    self.double_click_erase(zoom_corrected_pos)

        self.vdp.mouseDoubleClickEvent(event)

    def dialog_fill_erase(self, action='fill'):
        choice = QMessageBox.question(self, 'Double click detected!',
                                      'Do you want to '+ action +' the shape?',
                                      QMessageBox.Yes | QMessageBox.No)
        if choice == QMessageBox.Yes:
            return True
        else:
            return False


    def double_click_fill(self, zoom_corrected_pos):
        try:
            # erased, drawn = self.get_user_drawing(show_erased=True)
            usr_drawing = self.get_user_drawing()
            labs = label(usr_drawing, connectivity=2, background=0)
            # do I really need that can I rather do np.unique or alike and will it be faster
            rps_user_drawing = regionprops(labs)
            if len(rps_user_drawing) == 0:
                return
            drawn_mask = self.get_mask()
            # cell clicked over
            cell_id_clicked_over = labs[int(zoom_corrected_pos.y()), int(zoom_corrected_pos.x())]
            if cell_id_clicked_over == 0:
                return
            # we remove the cell drawn over
            drawn_mask[labs==cell_id_clicked_over] = 0
            # maintenant j'inverse et je trouve la cellule à filler sur le negatif
            lab_cells_user_drawing = label(drawn_mask, connectivity=1, background=255)
            cell_id_clicked_over2 = lab_cells_user_drawing[int(zoom_corrected_pos.y()), int(zoom_corrected_pos.x())]
            drawn_mask[lab_cells_user_drawing==cell_id_clicked_over2]=255
            self.set_mask(drawn_mask)
        except:
            logger.error('Double click fill failed')
            traceback.print_exc()


    def double_click_erase(self, zoom_corrected_pos):
        try:
            # erased, drawn = self.get_user_drawing(show_erased=True)
            erased, drawn = self.get_user_drawing(show_erased=True)
            labs = label(erased, connectivity=2, background=0)

            # do I really need that can I rather do np.unique or alike and will it be faster
            rps_user_drawing = regionprops(labs)
            if len(rps_user_drawing) == 0:
                return
            drawn_mask = self.get_mask()
            # cell clicked over
            cell_id_clicked_over = labs[int(zoom_corrected_pos.y()), int(zoom_corrected_pos.x())]
            if cell_id_clicked_over==0:
                return
            # we refill the removed cell
            drawn_mask[labs==cell_id_clicked_over] = 255
            # maintenant j'inverse et je trouve la cellule à filler sur le negatif
            lab_cells_user_drawing = label(drawn_mask, connectivity=1, background=0)
            cell_id_clicked_over2 = lab_cells_user_drawing[int(zoom_corrected_pos.y()), int(zoom_corrected_pos.x())]
            drawn_mask[lab_cells_user_drawing==cell_id_clicked_over2]=0


            self.set_mask(drawn_mask)
        except:
            logger.error('Double click delete failed')
            traceback.print_exc()


    def paintEvent(self, event):
        # super(Createpaintwidget, self).paintEvent(event) # KEEP somehow tjis causes bugs --> do not uncomment it
        canvasPainter = QtGui.QPainter(self)
        try:

            # the scrollpane visible region
            try:
                visibleRegion = self.scrollArea.widget().visibleRegion()
            except:
                #assume no scroll region --> visible region is self visible region
                visibleRegion = self.visibleRegion()
            # the corresponding rect
            visibleRect = visibleRegion.boundingRect()
            # the visibleRect taking zoom into account
            scaledVisibleRect = QRect(visibleRect.x() / self.scale, visibleRect.y() / self.scale,
                                      visibleRect.width() / self.scale, visibleRect.height() / self.scale)

            if self.image is None:
                canvasPainter.eraseRect(visibleRect)
                # canvasPainter.end()
                return

            canvasPainter.drawImage(visibleRect, self.image, scaledVisibleRect)
            if not self.vdp.active and self.maskVisible and self.imageDraw is not None:
                canvasPainter.drawImage(visibleRect, self.imageDraw, scaledVisibleRect)
                # should draw the cursor
            canvasPainter.drawImage(visibleRect, self.cursor, scaledVisibleRect)

            # ça ajoute un texte --> pas mal mais le refresh on scroll est pas top --> dois je le dessiner plutot dans vdp --> another set of objects that are non interactible and also cannot be saved and local???

            # if True:
            #     print('adding text')
            #     myQPen = QPen(QtCore.Qt.blue, 3)
            #     canvasPainter.setPen(myQPen)
            #     canvasPainter.setPen(QColor("Green"))
            #     canvasPainter.setFont(QFont('Helvetica', 48))
            #     #     print('drawing text')
            #     canvasPainter.drawText(visibleRect.x(), visibleRect.y() +10, 'this is a test of your system')
            #     # pb here is that the images is not cleared --> creates bugs

            if self.vdp.active:
                self.vdp.paintEvent(canvasPainter, scaledVisibleRect)
        except:
            traceback.print_exc()
        finally:
            canvasPainter.end()

if __name__ == '__main__':
    # TODO add a main method so it can be called directly
    # maybe just show a canvas and give it interesting props --> TODO --> really need fix that too!!!
    import sys
    from qtpy.QtWidgets import QApplication
    from matplotlib import pyplot as plt

    # should probably have his own scroll bar embedded somewhere

    app = QApplication(sys.argv)


    # in fact that is easy to override methods before construction --> can easily do that in fact

    # quite easy todo in fact
    class overriding_apply(Createpaintwidget):
        def apply(self):
            plt.imshow(self.get_mask())
            plt.show()

    #Do not use show() in a GUI application.
    # If you have a FigureCanvas instance embedded in your app, call its
    # draw() method.
    # If you use pyplot.figure() to create a matplotlib window from your
    # app, call pyplot.draw().

    # maybe add a figure canvas the to my code and plot to it and draw it --> quite easy and nice plotting capabilities then in python



    # w = Createpaintwidget(enable_shortcuts=True)
    w = overriding_apply(enable_shortcuts=True)
    w.set_image('/E/Sample_images/sample_images_PA/trash_test_mem/mini (copie)/focused_Series012.png')
    # w.set_image('/E/Sample_images/fluorescent_wings_spots_charroux/909dsRed/0.tif')
    # w.set_image('/E/Sample_images/fluorescent_wings_spots_charroux/contoles_test_X1VK06/0.tif')
    # w.set_mask('/E/Sample_images/sample_images_PA/trash_test_mem/mini (copie)/focused_Series012/handCorrection.png')
    # w.set_image(Img('/E/Sample_images/sample_images_PA/trash_test_mem/mini (copie)/focused_Series012.png'))

    # print(tst.max())
    # tst = tst[..., np.newaxis]




    # tst = Img('/E/Sample_images/fluorescent_wings_spots_charroux/909dsRed/0.tif')
    # tst = Img('/E/Sample_images/fluorescent_wings_spots_charroux/contoles_test_X1VK06/0.tif')
    # tst = Img('/E/Sample_images/sample_images_PA/trash_test_mem/mini (copie)/focused_Series012.png')
    # tst = Img.normalization(tst, method='Rescaling (min-max normalization)', range=[0,1], individual_channels=len(tst.shape)>2, clip=True)
    # tst = Img.normalization(tst, method='Rescaling (min-max normalization)', range=[0,1], individual_channels=len(tst.shape)>2, clip=True)
    # tst = Img.normalization(tst, method='Standardization', range=[0,1], individual_channels=True, clip=True)
    # tst = Img.normalization(tst, method='Percentile', range=[0,1], individual_channels=len(tst.shape)>2, clip=True, normalization_minima_and_maxima=[1.,50.0])
    # tst = np.squeeze(tst)

    # tst = auto_threshold(tst)
    # save_as_tiff(tst, output_name='/E/Sample_images/sample_images_PA/trash_test_mem/mini (copie)/tst_norm.tif')

    # do I have a standar
    # tst = Img.normalization(tst, method='standardization', range=[0,1], individual_channels=True)
    # print(tst.max())
    # w.set_image(tst)

    # all is so easy to do this way


    # ça marche --> tres facile a tester
    # --> how can I do a zoom
    # ask whether self scroll or not --> if scrollable

    # mask = w.get_mask()
    #
    # plt.imshow(mask)
    # plt.show()

    w.show()
    sys.exit(app.exec_())




# not so bad --> see how to add legends --> not so easy because I need to create various groups
# the ideal situtation would be to preview all the possible fusions and ask the user what should be done
# or offer a preview if user wants it and then ask him what to do
# otherwise fuse things
# NB add row names and column names to each image and offer the possibility to show or fuse them whenever necessary... --> will replace in an easier way the add labels around ...  --> think if smart or not ???
# do the math also below
# see ho to handle row labels
# preview might be slow --> ask how to do things ????


# https://docs.python.org/2/library/operator.html

# TODO try remove row if possible to only handle panels
# in a way a figure can be a panel of panels --> if so the I would just have one element --> in fact that would really simplify things like crazy
# pas mal mais du coup faudrait pas packer par defaut ou bien si sauf si l'utilisateur veut pas
# peut être plus flexible --> supress fig and row just keep panel and make it more flexible
# faire un stacker(horizontal or vertical)
# that may work in a way
# seule difference est que le panel ne doit contenir que des images de même taille alors que la row non mais en gros la row est un panel de panel
# puis je appliquer la regle des rows aux panels
# --> j'ai besoin d'une orientation des le depart
# est ce que ça peut marcher --> peut etre
# should I put a warning if they don't have the same size initially --> ? think about it
# should I have linear panel and 2D panel and give them different rules panel1D panel2D #dans panel2D ttes les images doivent avoir la meme taille. Panel1D --> c'est un peu une row en fait

# panel1D --> orientation --> hor/ver
# panel2D --> more stuff and same size required
import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
from qtpy import QtWidgets

from epyseg.draw.shapes.image2d import Image2D
from epyseg.draw.shapes.point2d import Point2D
from epyseg.draw.shapes.rect2d import Rect2D
from qtpy.QtCore import QPointF, QRectF
# logger
# from epyseg.ezfig import MyWidget
# from epyseg.figure import fig_tools
from epyseg.figure.alignment import alignTop, packX, updateBoudingRect, setToHeight, _brute_force_find_height
from epyseg.figure.fig_tools import preview, get_master_bounds2, get_common_scaling_factor
from epyseg.tools.logger import TA_logger

logger = TA_logger()

# keep this in mind: https://stackoverflow.com/questions/34827132/changing-the-order-of-operation-for-add-mul-etc-methods-in-a-custom-c
# do rows and columns and no need to take images --> they just pack horiz or vertically and that's it each image or stuff handles itslef
# a row like that cannot be appended another row or can it --> yes in fact just add to it
# not a good idea to have rect2D and qrectF at the same time....
# Ideally choose one or allow a better sync between the two

class Row(Rect2D):

    def __init__(self, *args, space=3, width=None):
        super(QRectF, self).__init__()
        self.images = []
        for img in args:
            # if isinstance(img, Row):
            #     self.images += img.images
            # else:
            # prevent adding images that could not be loaded
            if img.isSet:
                self.images.append(img)
        # init super empty
        self.isSet = True  # required to allow dragging
        self.space = space
        # why did not work ???
        self.sameHeight(space=space)
        self.widthInPixel = width
        self.setToWidth(width)

    # maybe also make the add a

    def __mod__(self, other):
        # formerly was a swap but in fact need better for a swap maybe == or <>
        if isinstance(other, Row):
            pos = self.get_P1()
            other_pos = other.get_P1()
            self.set_P1(other_pos)
            other.set_P1(pos)
        else:
            logger.error('swapping not implemented yet for ' + str(type(other)))

    def __ior__(self, other):
        return self.__or__(other=other)

    def __or__(self, other):
        # print('called')
        return self.__add__(other=other)
        # formerly was a swap but in fact need better for a swap maybe == or <> moved to modulo for now
        # if isinstance(other, Row):
        #     pos = self.get_P1()
        #     other_pos = other.get_P1()
        #     self.set_P1(other_pos)
        #     other.set_P1(pos)
        # else:
        #     logger.error('swapping not implemented yet for ' + str(type(other)))

    def __add__(self, other):
        # print('here', other)
        # from epyseg.figure.panel import Panel
        if isinstance(other, Row):
            # print(self.images)
            # print(len(self.images))
            # print(other.images)
            # print(len(other.images))
            P1 = self.get_P1()
            final_width = self.width()
            self.images = self.images + other.images

            # self.sameHeight(space=self.space)
            # self.setToHeight(self.height())
            self.setToWidth(final_width)
            self.set_P1(P1)
        # elif isinstance(other, Image2D):
        else:
            P1 = self.get_P1()
            final_width = self.width()
            self.images.append(other)
            # self.sameHeight(space=self.space)
            # self.setToHeight(self.height())
            self.setToWidth(final_width)
            self.set_P1(P1)
        # elif isinstance(other, Panel):
        #     self.images.append(other)
        #     self.sameHeight(space=self.space)
        #     self.setToWidth(self.widthInPixel)


        return self

        # def __truediv__(self, other):

    def __floordiv__(self, other):
        from epyseg.figure.column import Column
        P1 = self.get_P1()
        # final_height = self.height()
        col = Column(self, other,
                     space=self.space)  # the size of this stuff is way too big and way bigger than its components --> do not allow that
        # col.setToHeight(final_height)
        # need update the bounding rect
        col.sameWidth(self.space)
        # row.setToWidth(self.widthInPixel)
        col.set_P1(P1)
        return col

    # TODO maybe make it smarter so that it also removes from lower panels
    def __sub__(self, other):
        if other in self.images:
            self.images.remove(other)
        self.setToWidth(self.widthInPixel)
        return self

    def is_empty(self):
        if self.images is None or not self.images:
            return True
        return False

    def __truediv__(self, other):
        return self.__floordiv__(other=other)

    def __len__(self):
        if self.images is None:
            return 0
        return len(self.images)

    # def setWidthInPixel(self, widthInPixel):
    #     self.widthInPixel = widthInPixel
    #     self.packX()

    def packX(self, space=3):
        last_x = 0
        last_y = 0

        for i in range(len(self.images)):
            img = self.images[i]
            if i != 0:
                last_x += space
            img.set_P1(last_x, img.get_P1().y())
            last_x = img.boundingRect().x() + img.boundingRect().width()

        self.updateBoudingRect()

    def packY(self, space=3):
        last_x = 0
        last_y = 0

        for i in range(len(self.images)):
            img = self.images[i]
            if i != 0:
                last_y += space
            img.set_P1(img.get_P1().x(), last_y)
            # get all the bounding boxes and pack them with desired space in between
            # get first point and last point in x
            x = img.boundingRect().x()
            y = img.boundingRect().y()
            last_x = img.boundingRect().x() + img.boundingRect().width()
            last_y = img.boundingRect().y() + img.boundingRect().height()
        self.updateBoudingRect()

    def updateBoudingRect(self):
        '''updates the image bounding rect depending on content'''
        x = None
        y = None
        x2 = None
        y2 = None
        for img in self:

            # topLeft = img.get_P1()
            topLeft = img.boundingRect().topLeft()
            # print('comparison', img.boundingRect(), img.get_P1(), img) # the bug is here the various objects are not in sync !!!! --> do not allow that

            if x is None:
                x = topLeft.x()
            if y is None:
                y = topLeft.y()
            x = min(topLeft.x(), x)
            y = min(topLeft.y(), y)

            # print(img, img.boundingRect(), type(img))
            # print(img, img.boundingRect(), type(img), img.boundingRect().height())
            if x2 is None:
                x2 = topLeft.x() + img.boundingRect().width()
            if y2 is None:
                y2 = topLeft.y() + img.boundingRect().height()
            x2 = max(topLeft.x() + img.boundingRect().width(), x2)
            y2 = max(topLeft.y() + img.boundingRect().height(), y2)

        self.setX(x)
        self.setY(y)
        self.setWidth(x2 - x)
        self.setHeight(y2 - y)

        # print('bounds from inside', self)
        # bounds = get_master_bounds2(self.images)
        # self.setX(bounds.x())
        # self.setY(bounds.y())
        # self.setWidth(bounds.width())
        # self.setHeight(bounds.height())
        # return bounds

    # def boundingRect(self):
    #     return self.updateBoudingRect()
    # return get_master_bounds2(self.images)

    def setOrigin(self, *args):
        self.set_P1(*args)

    def set_P1(self, *args):
        curP1 = self.get_P1()
        Rect2D.set_P1(self, *args)  # is this how you call super ???
        newP1 = self.get_P1()
        for img in self:
            img.translate(newP1.x() - curP1.x(), newP1.y() - curP1.y())

        self.updateBoudingRect()

    def translate(self, *args):
        if len(args) == 1:
            point = args[0]
            QRectF.translate(self, point.x(), point.y())
            for img in self:
                img.translate(point.x(), point.y())
        else:
            QRectF.translate(self, args[0], args[1])
            for img in self:
                img.translate(QPointF(args[0], args[1]))

    def draw(self, painter, draw=True):
        if draw:
            painter.save()
        if draw:
            for img in self:
                img.draw(painter, draw=draw)
            painter.restore()

    def fill(self, painter, draw=True):
        if self.fill_color is None:
            return
        if draw:
            painter.save()
        if draw:
            for img in self:
                img.fill(painter, draw=draw)
            painter.restore()

    def drawAndFill(self, painter):
        painter.save()
        self.draw(painter, draw=True)
        painter.restore()

    def __iter__(self):
        ''' Returns the Iterator object '''
        self._index = -1
        return self

    def __next__(self):
        '''' Returns the next image or panel in the row '''
        self._index += 1
        if self._index < len(self.images):
            return self.images[self._index]
        # End of Iteration
        raise StopIteration

    def __lshift__(self, other):
        # move image left with respect to self
        if isinstance(other, Image2D):
            # swap the position of the two images and repack
            if other in self.images:
                pos = self.images.index(other)
                if pos - 1 >= 0:
                    self.images[pos - 1], self.images[pos] = self.images[pos], self.images[pos - 1]
                else:
                    return self
                self.packX(self.space)
                return self
        else:
            logger.error('not implemented yet swapping two objects ' + str(type(other)))

    def __rshift__(self, other):
        # move image left with respect to self
        if isinstance(other, Image2D):
            # swap the position of the two images and repack
            if other in self.images:
                pos = self.images.index(other)
                if pos + 1 < len(self.images):
                    self.images[pos + 1], self.images[pos] = self.images[pos], self.images[pos + 1]
                else:
                    return self
                self.packX(self.space)
                return self
        else:
            logger.error('not implemented yet swapping two objects ' + str(type(other)))

    # en fait ne peut pas marcher faut faire row per row
    def sameHeight(self, space):
        if space is None:
            space = 0
        max_height = None
        for img in self:
            if max_height is None:
                # print('TUTU',img)
                # print('TUTA',img, img.boundingRect(), img.boundingRect().height)
                # print('TOTO',img, img.boundingRect(), img.boundingRect().height())
                max_height = img.boundingRect().height()
            max_height = max(img.boundingRect().height(), max_height)
        for img in self:
            img.setToHeight(max_height)
            # print('height after update', img.height())

        self.packX(space)
        self.alignTop(updateBounds=False)
        self.updateBoudingRect()

    # Aligns objects within the block
    def arrangeRow(self):
        self.alignTop()

    # Align vectorial objects to the top
    # in fact should align in y
    def alignTop(self, updateBounds=False):
        first_left = None
        for img in self:
            cur_pos = img.get_P1()
            if first_left is None:
                first_left = cur_pos
            img.set_P1(cur_pos.x(), first_left.y())
        if updateBounds:
            self.updateMasterRect()

    def get_shape_at_coord(self, x, y):
        '''returns the shape under the mouse or None if none is found

        :param x:
        :param y:
        :return:
        '''
        for img in self.images:
            if img.boundingRect().contains(x, y):
                return img
        return None

    # ça marche super mais cree un bug pour les objects vectoriels --> voir où est le bug dans le truc
    # much smarter and mathematically correct set to width
    # Forces the block to be of width (width_in_px)
    def setToWidth(self, width_in_px):
        # from timeit import default_timer as timer
        # start = timer()
        if width_in_px is None:
            return

        topleft = Point2D(self.boundingRect().topLeft())
        self.sameHeight(self.space)

        # print('setToWidth row 1', timer() - start)

        # print('setToWidth row 2', timer() - start)
        # alignTop(topleft, *self.images)
        # print('setToWidth row 3', timer() - start)
        # packX(self.space, topleft, *self.images)

        # self.updateBoudingRect()  # I should also updat the bounding rect of output -> that is my error

        sizes = []
        for elm in self:
            sizes.append(elm.width())
            # print("inner size", elm)

        # new way to compute size --> much smarter than previous way --> start replace everywhere
        # sizes
        scaling = get_common_scaling_factor(sizes, width_in_px, self.space)
        # print('scaling, ', scaling)
        # need set to scale
        # need

        for elm in self:
            # elm.setWidth(scaling*elm.width())
            # elm.setHeight(scaling*elm.height())
            try:
                elm.scale *= scaling
            except:
                # scaling will not be working if elm has incompressible part --> need change scale again
                # c'est pas ce que je veux en fait --> c'est vraiment le mettre à la width desiree en fait
                elm.setToWidth(elm.width() * scaling)

        # pb there is a bug and the image and stuff don't match definitely my new code is better and smarter
        # this does noyt work but why ????
        # need pack them in x
        packX(self.space, None, *self.images)

        # need check that it worked

        # self.updateBoudingRect()
        # in fact that does not work and solving the equations can be difficult when the initial guess is wrong --> so brute force find it


        min_h = 10000000
        max_h = 0
        for elm in self:
            height = elm.height()
            # print('size of inner elm', )
            min_h = min(min_h, height)
            max_h = max(max_h, height)

        if max_h-min_h > 0.05:
            # closest = _brute_force_find_height(self, min_h - 1, max_h + 1, 1, width_in_px)

            # TODO --> need do something for gigantic images --> TODO
            # apparently this is painful with big images --> need a ratio
            # closest = _brute_force_find_height(self, min_h - 1, max_h + 1, (max_h - min_h) / 10, width_in_px)
            #
            # if closest is not None:
            #     if closest[1] - width_in_px < 0.05:
            #         self.setToHeight(closest[2])
            #         return closest
            # else:
            #     closest = _brute_force_find_height(self, min_h - 1, max_h + 1, 1, width_in_px)
            #
            # value = closest[1]
            # closest = _brute_force_find_height(self, value - 1, value + 1, 1, width_in_px)


            closest = _brute_force_find_height(self, min_h - 1, max_h + 1, 1, width_in_px)

            # print('closest1', closest)
            value = closest[1]
            # for height in np.arange(value - 1, value + 1, 0.01):
            #     col1.setToHeight(height)
            #     print('bob', height, col1.width(), col1.height())
            #     if (col1.width() - desired_width) <= mn and col1.width() - desired_width >= 0:
            #         closest = (col1.width(), col1.height(), height)
            #         mn = col1.width() - desired_width

            if closest[1]-width_in_px<0.05:
                self.setToHeight(closest[2])
                return closest

            closest = _brute_force_find_height(self, value - 0.5, value + 0.5, 0.1, width_in_px)


            # tres lent --> need skip steps if good enough
            # --> pas mal et assez rapide --> pourrait faire n iterations
            if closest[1]-width_in_px<0.05:
                self.setToHeight(closest[2])
                return closest

            value = closest[1]
            closest = _brute_force_find_height(self, value - 0.1, value + 0.1, 0.01, width_in_px)  # almost perfect

            # (128.00066196944778, 42.80999999999944, 42.80999999999944) --> solution est à peu pres ça --> comment la trouver
            # (128.5352647965376, 43.0, 43) # ineteger search then refinement --> could search between max height and min height of the shape of the column
            # (128.00066196944888, 42.80999999999984, 42.80999999999984)

            # this is a very good estimate

            self.setToHeight(closest[2])

        self.set_P1(topleft)
        self.updateBoudingRect()

    # def get_equa(self):
    #     sqdsqdsqdsqd

    # this is a really crappy method --> needs be changed
    # is that always correct irrespective of content ???
    # probably need to overwrite the set to scale of this
    # Forces the block to be of width (width_in_px)
    # def setToWidth_old(self, width_in_px):
    #     # from timeit import default_timer as timer
    #     # start = timer()
    #     if width_in_px is None:
    #         return
    #     # print('setToWidth row 1', timer() - start)
    #     topleft = Point2D(self.boundingRect().topLeft())
    #     # print('setToWidth row 2', timer() - start)
    #     alignTop(topleft, *self.images)
    #     # print('setToWidth row 3', timer() - start)
    #     packX(self.space, topleft, *self.images)
    #     # print('setToWidth row 4', timer() - start)
    #     # print('bounds before', updateBoudingRect(*self.images))
    #     setToWidth(self.space, width_in_px, *self.images)
    #     # print('setToWidth row 5', timer() - start)
    #     bounds = updateBoudingRect(*self.images)
    #     # print('bounds after', updateBoudingRect(*self.images))
    #     # print('setToWidth row 6', timer() - start)
    #
    #     max_height = -1
    #
    #     cur_height = bounds.height()
    #     cur_width = bounds.width()
    #     all_heights_are_the_same = True
    #     # print('setToWidth row 7', timer() - start)
    #     for img in self.images:
    #         max_height = max(max_height, img.boundingRect().height())
    #         # print('cur height', img.boundingRect().height(), img.boundingRect().width())
    #         if cur_height != max_height:
    #             all_heights_are_the_same = False
    #
    #     if not all_heights_are_the_same:
    #         setToHeight(self.space, max_height, *self.images)
    #         bounds = updateBoudingRect(*self.images)
    #         all_heights_are_the_same = True
    #     else:
    #         bounds = updateBoudingRect(*self.images)
    #
    #     sign = 1
    #     if bounds.width() >= width_in_px:
    #         sign = -1
    #
    #     # print(bounds.width(), 'vs desired', width_in_px, 'sign', sign)
    #     #
    #     # print(all_heights_are_the_same)
    #     # print(bounds.width() != width_in_px)
    #     # print(abs(bounds.width()-width_in_px)>0.3)
    #
    #
    #     # print('setToWidth row 8', timer() - start)
    #     # dirty height/width fix for complex panels figure out the math for it and replace this code
    #     # TODO also speed this up so that the action is only faked and not really done (just done once at the end to gain time)
    #
    #     # VERY DIRTY HERE --> need be hacked and replaced by new code
    #
    #     if not all_heights_are_the_same or (bounds.width() != width_in_px and abs(bounds.width()-width_in_px)>0.3):
    #
    #         # indeed this is super dirty code that need be replaced by the new code I have that doesn't use that
    #         # print('setToWidth row 9', timer() - start)
    #         while True:
    #             setToHeight(self.space, max_height, *self.images)
    #             bounds = updateBoudingRect(*self.images)
    #             if sign<0:
    #                 if bounds.width() <= width_in_px:
    #                     break
    #             else:
    #                 if bounds.width() >= width_in_px:
    #                     break
    #             max_height += 1 * sign
    #         # print('setToWidth row 10', timer() - start)
    #
    #         if bounds.width() != width_in_px:
    #             if bounds.width() >= width_in_px:
    #                 sign = -1
    #             else:
    #                 sign = 1
    #             # what if i remove refine step
    #             #
    #             # print('setToWidth row 10', timer() - start)
    #             # max_height -= 1
    #             # print('max_height', max_height, updateBoudingRect(*self.images))
    #             while True:
    #                 setToHeight(self.space, max_height, *self.images)
    #                 bounds = updateBoudingRect(*self.images)
    #                 if sign < 0:
    #                     if bounds.width() <= width_in_px:
    #                         break
    #                 else:
    #                     if bounds.width() >= width_in_px:
    #                         break
    #                 max_height += 0.25 *sign # critical step with this value --> timing is ok --> maybe ok for now
    #             # # print('setToWidth row 11', timer() - start)
    #
    #     # print('setToWidth row 11', timer() - start)
    #     # print('setToWidth row 12', timer() - start)
    #
    #
    #     self.updateBoudingRect() # I should also updat the bounding rect of output -> that is my error
    #
    #     # print('setToWidth row last', timer() - start)

    # def setToWidth(self, width_in_px):
    #     if width_in_px is None:
    #         return
    #     topleft = Point2D(self.boundingRect().topLeft())
    #     alignTop(topleft, *self.images)
    #     # because can use relative width and height
    #     # packX(self.space, topleft, *self.images)# do I need that --> I guess not
    #     # just fake it in fact --> faster
    #     fakeSetToWidth(width_in_px, *self.images)
    #     # compute fake nounds
    #     bounds = fakeUpdateBoudingRect(*self.images)
    #
    #
    #     # need compute fake height too
    #     min_height = 100_000_000
    #
    #     cur_height = bounds.height()
    #     cur_width = bounds.width()
    #     all_heights_are_the_same = True
    #     for img in self.images:
    #         min_height = min(min_height, img.boundingRect().height())
    #         if min_height<cur_height:
    #             all_heights_are_the_same=False
    #
    #     # dirty height/width fix for complex panels figure out the math for it and replace this code
    #     # TODO also speed this up so that the action is only faked and not really done (just done once at the end to gain time)
    #     if not all_heights_are_the_same or cur_width!=width_in_px:
    #         while True:
    #             setToHeight(min_height, *self.images)
    #             bounds = updateBoudingRect(*self.images)
    #             if bounds.width() >= width_in_px:
    #                 break
    #             min_height += 1
    #
    #         min_height -= 1
    #         while True:
    #             min_height += 0.05
    #             setToHeight(min_height, *self.images)
    #             bounds = updateBoudingRect(*self.images)
    #             if bounds.width() >= width_in_px:
    #                 break
    #
    #     self.updateBoudingRect()

    def setToHeight(self, height_in_px):
        if height_in_px is None:
            return
        # # nb_cols = len(self)
        # pure_image_height = (self.boundingRect().height()) - self.getIncompressibleHeight()
        # # print(width_in_px, self.getIncompressibleWidth(), (nb_cols - 1.) * self.space )
        # height_in_px -= self.getIncompressibleHeight()
        # ratio = height_in_px / pure_image_height
        # for img in self:
        #     img.setToHeight(img.boundingRect().height() * ratio)
        # self.packX(self.space)
        # self.updateBoudingRect()
        topleft = Point2D(self.boundingRect().topLeft())
        alignTop(topleft, *self.images)
        setToHeight(self.space, height_in_px, *self.images)
        self.updateBoudingRect()

    # @return the block incompressible width
    def getIncompressibleHeight(self):
        return 0

    # @return the block incompressible width
    def getIncompressibleWidth(self):
        if not self:
            return 0
        return (len(self) - 1.) * self.space

    # can I pass an ignoring char if some images need be ignored
    # probably need return the increment
    def setLettering(self, letter):
        # TODO implement increment
        # set letter for increasing stuff # check if letter is string or letter or array if so do or do not increase letters
        for img in self:
            img.setLettering(letter)
        # self.letter = letter

    # return all the solvable equations
    # all pairs of heights should be equal in a row
    # think how to do that
    # shall I actually build solvable equation sets ???
    def get_equation(self):
        from sympy import nsolve, exp, Symbol
        # TODO --> maybe get a width equation and a height one

        # need get the equa of inner also
        # maybe need even get the final equations to that I just append them all and get exactly what I need to solve
        # how should I increment that
        # UnevaluatedExpr(x + x)
        if not self:
            return None


        # in fact all heights should be equal and the sum of width would also be equal --> I need both sets of equations in fact --> maybe just do the get equation
        # maybe I should just append the equations
        # append both x and y equations and do the maths with them --> put them into a solvable form
        # what is the best way to do that ???
        equas_w = []
        equas_h = []
        equas_to_solve = []
        for img in self:
            equa_w, equa_h, equa_to_solve = img.get_equation()

            if equa_w:
                equas_w.extend(equa_w)
            if equa_h:
                equas_h.extend(equa_h)
            if equa_to_solve:
                equas_to_solve.extend(equa_to_solve)

            # do I need increment ??? --> maybe not or maybe yes
            # fig_tools.common_value+=1
        # do I need a formula there # that would be the sum of others
        # final equation is a concatenate of all equa for width
        # id(self) # to set the variable
        # --> is there an additional factor that would handle the resize if they are all the same size ???

        # in a row all heights need be the same --> must put the equas so that this is true
        # and in fact just return one equation for a row the top most row that needs be resized


        for iii in range(1, len(equas_h)):
            # print(iii)
            # first - sec = 0
            equa = equas_h[0] - equas_h[iii]
            equas_to_solve.append(equa)
        # if len(equa_h)==1:
        #     equa = equa_h[0] - equa_h[1]
        #     equas_to_solve.append(equa)

        final_equa = None
        for equa in equas_w:
            if final_equa is None:
                final_equa = equa
            else:
                final_equa += equa

        # see how I can do that
        # variable = Symbol('a'+str(id(self)))
        # final_equa

        # this is the equation I want
        final_equa += self.getIncompressibleWidth()
        equas_to_solve.append(final_equa) # pb is that for this equa I need to add the width --> how to I do that
        # this lists the variables in the formula --> do I don't need to store them separately
        # print(final_equa.free_symbols)  # returns all the variables contained in it

        # since they all have same height I need just return one equation
        return equas_w, [equas_h[0]], equas_to_solve

    # make no sense ??? la height n'est pas la somme mais juste une des valeurs de la colonne --> c'est tout je pense en fait
    # def get_height_equation(self):
    #     # TODO --> maybe get a width equation and a height one
    #
    #     # need get the equa of inner also
    #     # maybe need even get the final equations to that I just append them all and get exactly what I need to solve
    #     # how should I increment that
    #     # UnevaluatedExpr(x + x)
    #     equas = []
    #     for img in self:
    #         equas.append(img.get_height_equation())
    #         # do I need increment ??? --> maybe not or maybe yes
    #         # fig_tools.common_value+=1
    #     # do I need a formula there # that would be the sum of others
    #     # final equation is a concatenate of all equa for width
    #     # id(self) # to set the variable
    #     # --> is there an additional factor that would handle the resize if they are all the same size ???
    #     final_equa = None
    #     for equa in equas:
    #         if final_equa is None:
    #             final_equa = equa
    #         else:
    #             final_equa += equa
    #
    #     # see how I can do that
    #     # variable = Symbol('a'+str(id(self)))
    #     final_equa += self.getIncompressibleHeight() # useless because 0 but ok still
    #
    #     # this lists the variables in the formula --> do I don't need to store them separately
    #     print(final_equa.free_symbols)  # returns all the variables contained in it
    #
    #     return final_equa


    # # hack to make column hashable to be able to add it to a dict, see https://stackoverflow.com/questions/10994229/how-to-make-an-object-properly-hashable
    # def __eq__(self, other):
    #     try:
    #         return hash(str(self)) == other.__hash__()
    #     except:
    #         return False
    #
    # # hack to make column hashable to be able to add it to a dict, see https://stackoverflow.com/questions/10994229/how-to-make-an-object-properly-hashable
    # def __hash__(self):
    #     return hash(str(self))

    # allow to duplicate object
    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self

if __name__ == '__main__':
    img1 = Image2D('/E/Sample_images/sample_images_PA/trash_test_mem/counter/00.png')
    img2 = Image2D('/E/Sample_images/sample_images_PA/trash_test_mem/counter/01.png')
    img3 = Image2D('/E/Sample_images/sample_images_PA/trash_test_mem/counter/02.png')
    img4 = Image2D('/E/Sample_images/sample_images_PA/trash_test_mem/counter/03.png')
    # result = img1 + img2

    row = Row(img1, img2)
    row2 = Row(img3, img4)

    # peut etre la aussi le scripting language sera la clef
    # faire aussi un code de packing et de truc de ce style


    # print(result)
    result = row + row2
    print(result)
    result.sameHeight(3)
    result.setToWidth(128)

    print('equa', result.get_equation())
    # print('equa2', result.get_height_equation())

    print(result)

    # preview([result])
    preview(result)

    # not bad --> also do a
    # by default add all to the image
    # maybe have a 'list' command that can return the content of the IQL dict --> so that one can interact with them and one can also list the sub elements in EZFIG --> can be useful

    # TODO --> do a show of the stuff so that one can see the stuff

    # TODO --> do a plot of this
    # --> maybe need open an ezfig window --> TODO

    # find easy ways to letter and autoincrement sutff --> TODO
    # maybe store all as a script --> easy to reexecute --> but need store all images in the desired format



# will contain all tools for partial
from functools import partial

def get_method_name(partial_func):
    return partial_func.func.__name__

def is_partial(func):
    return isinstance(func, partial)

def are_function_names_matching(func1, func2):
    if is_partial(func1):
        func1 = get_method_name(func1)
    else:
        func1 = func1.__name__
    if is_partial(func2):
        func2 = get_method_name(func2)
    else:
        func2 = func2.__name__
    return func1 == func2

if __name__ == '__main__':
    def dummy():
        print('test')
    tst = partial(dummy)

    def dummy2():
        print('test2')

    print('is partial',is_partial(tst))

    print(get_method_name(tst))
    print(are_function_names_matching(dummy, dummy2))
    print(are_function_names_matching(dummy, dummy))
    print(are_function_names_matching(tst, dummy))


# cantor formula
# maybe the simplest way to give a unique id is that In mathematics, a pairing function is a process to uniquely encode two natural numbers into a single natural number. https://en.wikipedia.org/wiki/Pairing_function
# pairing function = ((x+y)*(x+y+1)/2)+y --> try that... --> but will that work with big ints as mine ??? and --> try with pure white and if ok then I have it # or use the local id to give a global id using this formula -->

# pi(k1, k2) = 1/2(k1 + k2)(k1 + k2 + 1) + k2
# pi(k1, k2) = 1/2(k1 + k2)(k1 + k2 + 1) + (k1*k2)


import numpy as np
from functools import partial
# from PIL import Image
from scipy.ndimage import generic_filter
from epyseg.img import Img, pad_border_xy
import matplotlib.pyplot as plt


# TODO have a look here to see if super fast or not https://stackoverflow.com/questions/10996769/pixel-neighbors-in-2d-array-image-using-python


# could create a file with vertices and bonds maybe with forbidden colors in fact
# ça y est j'ai les vertices exactement comme dans TA
# could ask for the color of the bonds --> can be black or white

# ask for what should be returned

# NB MEGA IMPORTANT IC OULD DEFINE BONDS BY VERTICES --> IF TWO DISTINCT VERTICES ARE ASSOCIATED TO THE SAME TWO CELL IT MEANS THEY ARE CONNECTED --> easy way of getting bonds and in addition also nicely detects bonds consisting of just two pixels --> smarter than what I used to do in TA!!! --> THINK ABOUT IT

# use the other method here because this is so slow


# MEGA TODO --> REPLACE TUPLE BY LIST for the coords of vertices !!!
def associate_vertices_to_cells(cell_labels, vertices, forbidden_colors=(0, 0xFFFFFF),
                                output_vertices_and_associated_cells=False, output_cells_and_their_vertices=False,
                                output_cells_and_their_neighbors=False):

    outputs = []
    # for all vertices detect the corresponding cell
    cell_id_n_vx_coords = {}
    # contains for each vertex all the associated cells --> can be used for rescue of neighbours for example or things alike
    vertices_n_associated_cells = {}

    if isinstance(vertices, tuple):
        vx_coords = vertices
    else:
        vx_coords = np.where(vertices == 255)  # assume vertices are labeled pure white...

    # print('dobs',type(vx_coords), type(vertices)) #dobs <class 'tuple'> <class 'numpy.ndarray'>
    # then need look around it and compute

    # print('x', vertices[0][0])
    # print('y', vertices[1][0])

    # look around each vx for ids and add the vx to all the corresponding ids
    # print(vx_coords[0], vx_coords[...,0])

    # the pb is that if vertices are missing definitely cells would be missing --> to avoid this I would need to get t
    # TODO FIX THIS TO DETECT ALL VERTICES (SEE THE TRIANGULATE CLASS FOR A FAILING EXAMPLE)!!!

    for i in range(len(vx_coords[0])):
        y = vx_coords[0][i]
        x = vx_coords[1][i]
        ids = neighbors8((y, x), cell_labels)
        # ids = neighbors8_2((y, x), cell_labels)

        # print('neighbs length', len(ids))

        ids = set(ids)  # --> ok but need remove pure white and black from that or ignore them # en fait je peux aussi faire ça...

        if forbidden_colors:
            for col in forbidden_colors:
                if col in ids:
                    ids.remove(col)

        # print(x,y, ids)
        # add all the neighbs of the current vx
        vertices_n_associated_cells[(y, x)] = ids
        # print('vx asssoc ids', ids)  #seems ok --> almost all 3

        for id in ids:
            if id in cell_id_n_vx_coords:
                # raw = cell_id_n_vx_coords[id]
                cell_id_n_vx_coords[id].append((y, x))
            else:
                cell_id_n_vx_coords[id] = [(y, x)]




    if output_cells_and_their_vertices:
        outputs.append(cell_id_n_vx_coords)

    if output_vertices_and_associated_cells:
        outputs.append(vertices_n_associated_cells)

    # TODO also offer the connexion of cells and all of its neighbors --> TODO
    # in fact I need bonds to get cells really connected or i need vertices

    if output_cells_and_their_neighbors:
        cell_ids_n_neighbor_ids = {}
        for key, value in cell_id_n_vx_coords.items():
            # print(key, '->', value)
            neighbs = []
            for vx in value:
                neighbs += vertices_n_associated_cells[vx]

            neighbs = set(neighbs)
            neighbs.remove(key)

            # if key == 8787781:
            #     print('souppa', len(neighbs), len(value), neighbs, value)

            cell_ids_n_neighbor_ids[key] = neighbs
        # print('final connection cell to cells', cell_ids_n_neighbor_ids)
        outputs.append(cell_ids_n_neighbor_ids)

    if len(outputs) == 1:
        return outputs[0]
    else:
        if outputs:
            return outputs

    # mega TODO --> return also the vertices and the cells --> offer this as an offer because can also be very useful and not very smart doing this several times because may be needed several times --> TODO
    return cell_id_n_vx_coords


# detects the neighbors of the lost cell so that I can reidentify the lost cells by their neighborhood
def detect_neighbors_of_current_cell(cell_id, cell_vertices, cell_labels, forbidden_colors=(0,0xFFFFFF)):
    # loop over cell vertices and detect neighbors
    # then find the cell that best matches this and assign it

    neighbs = []
    for vx in cell_vertices:

        # print('vx', vx)
        # cell_ids_n_neighbor_ids = {}
        # for key, value in cell_id_n_vx_coords.items():
            # print(key, '->', value)

            # for vx in value:

        # print('neighbors8', neighbors8(vx, cell_labels), neighbors8_2(vx, cell_labels)) # TODO replace method some day...
        neighbs += neighbors8(vx, cell_labels).tolist()

    neighbs = set(neighbs)
    neighbs.remove(cell_id)

    if forbidden_colors:
        for col in forbidden_colors:
            if col in neighbs:
                neighbs.remove(col)

            # if key == 8787781:
            #     print('souppa', len(neighbs), len(value), neighbs, value)

            # cell_ids_n_neighbor_ids[key] = neighbs
        # print('final connection cell to cells', cell_ids_n_neighbor_ids)
        # outputs.append(cell_ids_n_neighbor_ids)
    return neighbs

# eight_neighb = [[(-1,-1), (0,-1), (1,-1)],
#                 [(-1,0), 0, 1],
#                 [-1, 0, 1]]

def neighbors8(vx_coords, cell_labels):

    # print('vx_coords',vx_coords)

    min_y = vx_coords[0] - 1
    max_y = vx_coords[0] + 2
    min_x = vx_coords[1] - 1
    max_x = vx_coords[1] + 2

    min_y = min_y if min_y >= 0 else 0
    max_y = max_y if max_y < cell_labels.shape[0] else cell_labels.shape[0]
    min_x = min_x if min_x >= 0 else 0
    max_x = max_x if max_x < cell_labels.shape[1] else cell_labels.shape[1]

    return cell_labels[min_y:max_y, min_x:max_x].ravel()

def neighbors8_shift_mapping(vx_coords, cell_labels):
    min_y = vx_coords[0] - 1
    max_y = vx_coords[0] + 2
    min_x = vx_coords[1] - 1
    max_x = vx_coords[1] + 2

    # print('0min_x, max_x, min_y, max_y, vx_coords', min_x, max_x, min_y, max_y, vx_coords) # --> ok --> error in correc below

    min_y = min_y if min_y >= 0 else 0
    max_y = max_y if max_y < cell_labels.shape[0] else cell_labels.shape[0]
    min_x = min_x if min_x >= 0 else 0
    max_x = max_x if max_x < cell_labels.shape[1] else cell_labels.shape[1]

    # if min_x>max_x:
    #     min_x, max_x = max_x, min_x
    # if min_y>max_y:
    #     min_y, max_y = max_y, min_y

    # why not sorted
    # print('min_x, max_x, min_y, max_y, vx_coords',min_x, max_x, min_y, max_y, vx_coords)

    coords = []
    for y in range(min_y, max_y):
        for x in range(min_x, max_x):
                coords.append([y,x])

    return coords

    # [min_y:max_y, min_x:max_x]

    # return cell_labels[min_y:max_y, min_x:max_x].ravel()



# TODO compare both algos and find fastest
# DO NOT USE THIS ALGO AS IT IS MUCH SLOWER THAN THE PREVIOUS ONE!!!
# def neighbors8_2(vx_coords, cell_labels):
#     coords = []
#     for i in [-1, 0, 1]:
#         for j in [-1, 0, 1]:
#             if vx_coords[0]+i>=0 and vx_coords[0]+i<cell_labels.shape[0]:
#                 if vx_coords[1] + j >= 0 and vx_coords[1] + j < cell_labels.shape[1]:
#                     coords.append((vx_coords[0]+i, vx_coords[1]+j))
#     return coords
    # return cell_labels[min_y:max_y, min_x:max_x].ravel()



def cantor_color_bonds(RGB24_or_lab, vertices, boundary_color=0xFFFFFF,copy=True): #  # TODO use partial to send parameters to the method
    if isinstance(boundary_color, int):
        boundary_color = [boundary_color]
    method = partial(_cantor_color_bonds, boundary_colors=boundary_color) # bug if boundary color is set --> why --> most likely a bug in their code
    # vertices = generic_filter(RGB24, method, (3, 3))
    # return generic_filter(RGB24, method, (3, 3))

    # print(method)
    if copy:
        # copy image
        RGB24 = np.copy(RGB24_or_lab)
    else:
        # in place
        RGB24 = RGB24_or_lab
    # return generic_filter(RGB24, _cantor_color_bonds, (3, 3))

    # TODO handle border bonds that are wrong yet need be handled:
    # scan line by line and change id depending on


    bonds = generic_filter(RGB24, method, (3, 3))
    bonds[vertices != 0]=0


    # handle border bonds
    max_id = bonds.max()+1
    for iii in range(RGB24.shape[1]):
        if vertices[0,iii]==255:
            max_id+=1
            continue
        bonds[0,iii]=max_id
    for iii in range(RGB24.shape[1]):
        if vertices[RGB24.shape[0]-1,iii]==255:
            max_id+=1
            continue
        bonds[RGB24.shape[0]-1,iii]=max_id

    for jjj in range(RGB24.shape[0]):
        if vertices[jjj,0]==255:
            max_id+=1
            continue
        bonds[jjj,0]=max_id
    for jjj in range(RGB24.shape[0]):
        if vertices[jjj,RGB24.shape[1]-1]==255:
            max_id+=1
            continue
        bonds[jjj,RGB24.shape[1]-1]=max_id


    # return generic_filter(RGB24, method, (3, 3))
    return bonds


def detect_vertices_and_bonds(RGB24_or_lab, detect_bonds=False,
                              boundary_color=0xFFFFFF, split_bonds_and_vertices = False, copy=True, pad_to_detect_vertices=True):  # TODO use partial to send parameters to the method
    # print('inner shape',RGB24.shape) --> n'a pas marché

    if isinstance(boundary_color, int):
        boundary_color=[boundary_color]
    # method = partial(_identify_vertices_and_bonds, detect_bonds=detect_bonds)
    # return generic_filter(RGB24, _detect_vertices_and_bonds, (3, 3))
    # return generic_filter(RGB24, method, (3, 3))
    if copy and not pad_to_detect_vertices:
        # copy image
        RGB24= np.copy(RGB24_or_lab)
    else:
        # in place
        RGB24 = RGB24_or_lab


    # method = partial(_detect_vertices_and_bonds, detect_bonds=detect_bonds)
    method = partial(_detect_vertices_and_bonds, boundary_colors=boundary_color)
    # vertices = generic_filter(RGB24, method, (3, 3))

    if pad_to_detect_vertices:
        # do pad to get the right vertices --> TODO
        RGB24 = np.pad(RGB24, pad_width=1, mode='constant', constant_values=False)
        RGB24 = pad_border_xy(RGB24, mode=RGB24.max() + 1)
        # print(padded_cells.shape)

    # print('test', RGB24.shape, RGB24.dtype)
    vertices = generic_filter(RGB24, method, (3, 3))

    # up until here is ok

    if pad_to_detect_vertices:
        # add the few missing vertices to the paded image
        vertices = vertices[1:-1, 1:-1]  # unpad
        # add a vertex to each image corner
        vertices[0, 0] = 255
        vertices[-1, 0] = 255
        vertices[0, -1] = 255
        vertices[-1, -1] = 255

        RGB24 = RGB24[1:-1, 1:-1]# unpad RGB image too
    else:
        # old code, probably do not use because misses some vertices and is not very smart
        # detect vertices at image boundaries
        for jjj in range(1, RGB24.shape[0] - 1):
            neighbs = RGB24[jjj - 1:jjj + 2, 0:2]
            different_neighbours = set(neighbs.ravel().tolist())
            if len(different_neighbours) == 3:
                vertices[jjj][0] = 255
            neighbs = RGB24[jjj - 1:jjj + 2, RGB24.shape[0] - 2:RGB24.shape[0]]
            different_neighbours = set(neighbs.ravel().tolist())
            if len(different_neighbours) == 3:
                vertices[jjj][RGB24.shape[0] - 1] = 255

        for iii in range(1, RGB24.shape[1] - 1):
            neighbs = RGB24[0:2, iii - 1:iii + 2]
            different_neighbours = set(neighbs.ravel().tolist())
            if len(different_neighbours) == 3:
                vertices[0][iii] = 255
            neighbs = RGB24[RGB24.shape[1] - 2:RGB24.shape[1], iii - 1:iii + 2]
            different_neighbours = set(neighbs.ravel().tolist())
            if len(different_neighbours) == 3:
                vertices[RGB24.shape[0] - 1][iii] = 255

        # add a vertex to each image corner
        vertices[0][0] = 255
        vertices[RGB24.shape[0] - 1][0] = 255
        vertices[RGB24.shape[0] - 1][RGB24.shape[1] - 1] = 255
        vertices[0][RGB24.shape[1] - 1] = 255

    # NB TODO if detect bonds --> detect bonds at the edges too

    # fix bonds at the border of the image that are not detected
    if detect_bonds:
        # print('in here')
        # pb si trop de couleurs --> à fixer un jour en fait
        vertices[np.where((RGB24==boundary_color) & (vertices!=255))] = 128

    if split_bonds_and_vertices and detect_bonds:
        bonds = np.zeros_like(vertices)
        bonds[vertices == 128] = 128  # make a bond specific file
        vertices[vertices == 128] = 0  # make a vertex specific file
        return vertices, bonds




    return vertices


def _cantor_pairing(id1, id2, _sort_points=True):
    x = id1
    y = id2
    if _sort_points and id1<id2:
        x=id2
        y=id1
    # return ((x + y) * (x + y + 1) / 2) + y
    return int(((x + y) * (x + y + 1) / 2) + y) #



# def identify_bonds(RGB24):
#     # the idea here is to give to each bond a unique ID and also associate the vertices to each bond --> create the equivalent of the TA file --> TODO
#     # loop for every white pixel and if two ids around it give it a unique id
#     # if three or more ids it is a vertex --> could also give it an id maybe to be consitent and should associate it to some bonds if needed --> very close to the previous code but I would pass to this a dict on top of that where I would add the id
#
#     # try cantor formula to avoid having to store/pass a map and search it
#
#     pass

def _detect_vertices_and_bonds_old_doing_useless_stuff(P, detect_bonds=False, boundary_colors=(0xFFFFFF, 0)):
    if boundary_colors:
        if P[4] not in boundary_colors:  # not pure white
            return 0

    size = len(set(P))  # 4 way vertices only but why not
    if detect_bonds and size == 3:
        return 128
    # if size >= 4:
    #     print(set(P))
    return 255 if size >= 4 else 0

# this thing detects bonds and vertices but does not give an id to bonds
# TODO convert to vertices the things touching the boundary too ??? or not???
def _detect_vertices_and_bonds(P, boundary_colors=(0xFFFFFF, 0)):
    if boundary_colors:
        if P[4] not in boundary_colors:  # not pure white
            return 0

    size = len(set(P))  # 4 way vertices only but why not
    # if detect_bonds and size == 3:
    #     return 128
    # if size >= 4:
    #     print(set(P))
    return 255 if size >= 4 else 0


# unique_bond_identifiers --> would be a dict having an id and a corresponding pair of cells --> maybe sort them or add the entry twice --> maybe sorting is more efficient


# nb I could use np.unique to recolor all of the cells to something in the range 1-whatever --> good idea in fact and could be fast --> try that
def _identify_vertices_and_bonds(P, detect_bonds=False, boundary_colors=(0xFFFFFF, 0)):

    if boundary_colors:
        if P[4] not in boundary_colors:  # not pure white
            return 0

    corrected_ids = set(P)
    size = len(corrected_ids)  # 4 way vertices only but why not

    # ça a l'air de marcher
    if detect_bonds and size == 3:
        # print('in')
        corrected_ids = list(corrected_ids.difference(boundary_colors))
        # TODO use cantor pairing here to get a unique id from two bonds
        # print(corrected_ids[0], corrected_ids[1])
        # print(_cantor_pairing(corrected_ids[0], corrected_ids[1]))
        return _cantor_pairing(corrected_ids[0], corrected_ids[1])
        # return 128
    # if size >= 4:
    #     print(set(P))
    return 255 if size >= 4 else 0


def _cantor_color_bonds(P, boundary_colors=(0xFFFFFF, 0)):
    if boundary_colors:
        # print('in')
        if P[4] not in boundary_colors:  # not pure white
            return 0
    corrected_ids = set(P)
    size = len(corrected_ids)  # 4 way vertices only but why not
    if size == 3:
        corrected_ids = list(corrected_ids.difference(boundary_colors))
        return _cantor_pairing(corrected_ids[0], corrected_ids[1])
    return 0



if __name__ == '__main__':

    if False:
        import sys
        print(_cantor_pairing(0xFFFFFF,0xFFFFFE)) #562949382996104 562949886312449
        print(_cantor_pairing(0xFFFFFF,0xFFFFFF)) #562949382996104 562949886312449
        print(_cantor_pairing(0xFFFFFFFF,0xFFFFFFFE)) #3.6893488130239234e+19 --> very big
        print(_cantor_pairing(0xFFFFFFFF,0xFFFFFFFF)) #3.689348813882917e+19 --> very big too # try avoid such big number but ok if can't do better...
        print(_cantor_pairing(0, 0))
        print(_cantor_pairing(0, 1)) # 0,1 and 1, 0 differ --> need sort the points before passing them or after passing them
        print(_cantor_pairing(1, 0))
        print(_cantor_pairing(10, 23)) # ça marche variment bien mais faire +1 pr eviter d'avoir 0 --> en fait je m'en fous car pourra jamais etre = 0
        print(_cantor_pairing(23, 10))
        print(_cantor_pairing(1, 32))
        print(_cantor_pairing(32, 1))

        # --> I love this because this is new and elegant


        # very good idea
        # I love this cantor formula
        sys.exit(0)


    # Open image and make into Numpy array
    # PILim = Image.open('patches.png').convert('RGB')
    RGBim = Img('/E/Sample_images/segmentation_assistant/ovipo_uncropped/200709_armGFP_suz_46hAPF_ON.lif - Series011/tracked_cells_resized.tif')[0:64, 0:64]
    # RGBim = np.array(PILim)

    # Make a single channel 24-bit image rather than 3 channels of 8-bit each

    print(RGBim.shape)

    # there is a bug but almost there still
    RGB24 = (RGBim[..., 0].astype(np.uint32) << 16) | (RGBim[..., 1].astype(np.uint32) << 8) | RGBim[..., 2].astype(
        np.uint32)

    print(RGB24.max(), RGB24.min())
    # RGB24 = (RGBim[...,0].astype(np.uint32)<<16) | (RGBim[...,1].astype(np.uint32)<<8) | RGBim[...,2].astype(np.uint32)

    # Run generic filter counting unique colours in neighbourhood
    result = detect_vertices_and_bonds(RGB24)


    # then could recolor them if needed --> could try that
    # give a unique id to bonds --> very easy in facts --> maybe replace old code ??? --> pb is that I need a uint64 image that will be quite big in mem but ok I guess or need relabel the image afterwards --> maybe by the way...
    result = generic_filter(RGB24.astype(np.uint64), partial(_identify_vertices_and_bonds,detect_bonds=True), (3, 3))
    # try to see how I can do that better!!!


    # result = generic_filter(RGB24.astype(float), partial(_identify_vertices_and_bonds,detect_bonds=True), (3, 3))
    print(result.dtype)
    # result = result.astype(np.uint8)

    # Save result
    # Image.fromarray(result.astype(np.uint8)).save('result.png')
    plt.imshow(result)
    plt.show()

    # plt.imshow(RGB24)
    # plt.show()

    indices = (result == 255)
    print('indices', indices)

    vertices = np.where(result == 255) #c'est
    print(vertices)
    if vertices[0].size != 0:
        print('y', vertices[0][0])
    if vertices[1].size != 0:
        print('x', vertices[1][0])
    if vertices[0].size + vertices[1].size != 0:
        print('first vx', result[vertices[0][0], vertices[1][0]])

    print(associate_vertices_to_cells(RGB24,
                                      result))  # ça marche en fait --> j'ai associé les vertex à chaque cellule --> cool
    # pb need add vertices at the corners too --> like the first or second pixel away from the border



# tt est pas mal j'y suis presque mais juste demander ce qu'on doit sauver en fait et aussi juste sauver les weights --> TODO
# TODO add more models to come


# in fact --> can do a combinatorial test to see which combin is best
# all is ok
# need to do model testing --> the difficulty is the combination of denoiser and surface proj, th rest is easy
#


# here define the minimal code I need to do surface proj, with and without heightmap generation, etc...

# need load a model that contains surface proj and then need load a denoiser
# could also do the projection based on height map by taking average image
# --> TODO and should not be too hard in fact!!!

# TODO rather use the code in single predict as I can do more stuff and even extend the stack
# if single predict does not exist --> do it!!!!
import traceback

import matplotlib.pyplot as plt
import numpy as np
from epyseg.deeplearning.deepl import EZDeepLearning
import os
from epyseg.img import Img, has_metadata
from tensorflow.keras.layers import Lambda
from tensorflow.keras.models import Sequential

# img = Img('/E/Sample_images/sample_images_pyta/Image49.lsm')

# TODO --> maybe do a code that selects out of 10 models --> these models are anyway so small that maybe I can store everything on githib no pb
from epyseg.ta.database.sql import TAsql, _to_dict, update_db_properties_using_image_properties
from epyseg.ta.deep.create_model_for_projection import \
    create_surface_projection_denoise_and_height_map_combinatorial_model
from epyseg.tools.early_stopper_class import early_stop
from epyseg.ta.tracking.tools import smart_name_parser

# nb in the augs I could also invert the Z axis in fact not because that would lose height
from epyseg.ta.measurements.measurements3D.height_map_to_projection import surface_projection_from_height_map


# TODO I need to test that!!!

# TODO ADD CHANNEL SUPPORT HERE THEN DONE!!

# need get the images processed and to add them to the next list

# TODO --> put the recursion as a parameter and do a code to test all possible combinations to see if any would be better

# can almost reach the best score of CARE with that --> very good in fact --> do flip my flies
def surface_projection_pyta(deepTA, input_file, progress_callback=None, cur_progress=None, save_raw_image=False,
                            channel=None, recursion_for_denoising=1): # 4 ou 5 --> super good for CARE crappy images # --> 4 is good for CARE with super crappy images  one recursion is better for the other --> only makes sens to increase the nb of recursions if image is noisy otherwise ignore
    if isinstance(input_file, list):
        processed_files = []
        for lll, file in enumerate(input_file):
            cur_progress = (lll / len(input_file)) * 100
            try:
                if early_stop.stop == True:
                    return processed_files
                result = surface_projection_pyta(deepTA, file, progress_callback=progress_callback,
                                                 cur_progress=cur_progress, save_raw_image=save_raw_image,
                                                 channel=channel, recursion_for_denoising=recursion_for_denoising)
                if result is not None:
                    processed_files.append(result)
            except:
                traceback.print_exc()
                print('Error: could not process file: "' + str(file) + '"')
        return processed_files
    parent, filename = smart_name_parser(
        input_file,
        ordered_output=['parent', 'short_no_ext'])

    if early_stop.stop == True:
        return

    if progress_callback is not None and cur_progress is not None:
        progress_callback.emit(cur_progress)

    Z_FRAMES_TO_ADD = None  # 5 #None #5 --> often None is better --> much better quality than the others --> but not always true!!! --> in fact need test all the images directly without adding extra frames in fact!!!
    # --> so my ensemble testing sucks
    # ideally always pass in the model with height map and surface proj, so that the output results are awlays the same
    result = predict_single_image(deepTA, input_file, Z_FRAMES_TO_ADD=Z_FRAMES_TO_ADD,
                                  input_normalization={'method': Img.normalization_methods[7], 'range': [2, 99.8],
                                                       'individual_channels': True, 'clip': False})


    # could then repredict based on a smaller model

    surface_proj_file_path = os.path.join(os.path.join(parent, 'surface_projection'), filename + '.tif')

    # is there really a difference or am I doing smthg wrong ???
    # plt.imshow(result[0])
    # plt.show()
    # plt.imshow(result[1])
    # plt.show()

    if not save_raw_image:
        # old
        # denoised_surface_proj = result[0]
        # Img(np.squeeze(denoised_surface_proj), dimensions='hw').save(surface_proj_file_path)

        raw_surface_proj = result[1]
        # reprocess it

        Img(np.squeeze(raw_surface_proj), dimensions='hw').save(surface_proj_file_path)

        model = deepTA.model
        try:
            # very dirty hack to apply normalization to surface proj before running the denoiser again --> totally crap and sub optimal but kinda working --> try improve that some day!!!
            tmp_model = Sequential()# get the sub model here
            tmp_model.add(model.layers[-3])
            tmp_model.add(Lambda(lambda x: x[..., 0:1]))
            deepTA.model = tmp_model
            for rrr in range(recursion_for_denoising):
                result2 = predict_single_image(deepTA, surface_proj_file_path, Z_FRAMES_TO_ADD=Z_FRAMES_TO_ADD, input_channel_of_interest=None,
                                              input_normalization={'method': Img.normalization_methods[7], 'range': [2, 99.8],
                                                                   'individual_channels': True, 'clip': False})
                denoised = result2[0]
                Img(np.squeeze(denoised), dimensions='hw').save(surface_proj_file_path)
                del result2
                del denoised
            del tmp_model
            # print('result[0]', result2[0], result2              )
        except:
            traceback.print_exc()
            print('error denoising')
        finally:
            deepTA.model = model

        # then reprocess this image with the submodel and save it


        # c'est aussi là que je dois creer ou bien mettre à jour la db de TA idealment sans casser les colonnes existantes
    else:
        raw_surface_proj = result[1]
        # get parent dir and save there in the appropriate folder
        # surface_proj_file_path = os.path.join(os.path.join(parent,'surface_projection'), filename+'.raw.tif')

        # print(surface_proj.shape, surface_proj.dtype, surface_proj_file_path)
        Img(np.squeeze(raw_surface_proj), dimensions='hw').save(surface_proj_file_path)
        # ta_path = smart_name_parser(surface_proj_file_path,ordered_output=['TA'])

    ta_path = smart_name_parser(surface_proj_file_path, ordered_output='TA')

    height_map = result[2]

    # TODO compute heigh map quality (if out of frame --> need try yet another surface projection algorithm)
    # lui faire computer et sauver ce fichier

    if Z_FRAMES_TO_ADD is not None:
        height_map = height_map - Z_FRAMES_TO_ADD  # correct height map for added black frames --> then ideally I need also compute the heightmap quality that is
    Img(np.squeeze(height_map), dimensions='hw').save(os.path.join(ta_path, 'height_map.tif'))

    # rewind the stuff and do proj
    height_map_quality_test = surface_projection_from_height_map(Img(input_file), np.squeeze(height_map),
                                                                 channel=channel)  # TODO would need hack this to support the channel to be read for all images in case they are multichannel
    Img(height_map_quality_test, dimensions='hw').save(os.path.join(ta_path, 'height_map_quality_test.tif'))

    # save parent image properties in the db in the properties table so that they can be easily reused by TA
    update_db_properties_using_image_properties(input_file, ta_path)

    return surface_proj_file_path  # this is the file that needs be loaded in the other


def surface_projection_pyta_bckup(deepTA, input_file, progress_callback=None, cur_progress=None, save_raw_image=False,
                            channel=None):
    if isinstance(input_file, list):
        processed_files = []
        for lll, file in enumerate(input_file):
            cur_progress = (lll / len(input_file)) * 100
            try:
                if early_stop.stop == True:
                    return processed_files
                result = surface_projection_pyta(deepTA, file, progress_callback=progress_callback,
                                                 cur_progress=cur_progress, save_raw_image=save_raw_image,
                                                 channel=channel)
                if result is not None:
                    processed_files.append(result)
            except:
                traceback.print_exc()
                print('Error: could not process file: "' + str(file) + '"')
        return processed_files
    parent, filename = smart_name_parser(
        input_file,
        ordered_output=['parent', 'short_no_ext'])

    if early_stop.stop == True:
        return

    if progress_callback is not None and cur_progress is not None:
        progress_callback.emit(cur_progress)

    Z_FRAMES_TO_ADD = None  # 5 #None #5 --> often None is better --> much better quality than the others --> but not always true!!! --> in fact need test all the images directly without adding extra frames in fact!!!
    # --> so my ensemble testing sucks
    # ideally always pass in the model with height map and surface proj, so that the output results are awlays the same
    result = predict_single_image(deepTA, input_file, Z_FRAMES_TO_ADD=Z_FRAMES_TO_ADD,
                                  input_normalization={'method': Img.normalization_methods[7], 'range': [2, 99.8],
                                                       'individual_channels': True, 'clip': False})

    # could then repredict based on a smaller model

    surface_proj_file_path = os.path.join(os.path.join(parent, 'surface_projection'), filename + '.tif')

    # is there really a difference or am I doing smthg wrong ???
    plt.imshow(result[0])
    plt.show()
    plt.imshow(result[1])
    plt.show()

    if not save_raw_image:
        denoised_surface_proj = result[0]
        # get parent dir and save there in the appropriate folder

        # print(surface_proj.shape, surface_proj.dtype, surface_proj_file_path)

        # could take the raw image and reprocess it separately with a subset of the model --> TODO

        Img(np.squeeze(denoised_surface_proj), dimensions='hw').save(surface_proj_file_path)

        # c'est aussi là que je dois creer ou bien mettre à jour la db de TA idealment sans casser les colonnes existantes
    else:
        raw_surface_proj = result[1]
        # get parent dir and save there in the appropriate folder
        # surface_proj_file_path = os.path.join(os.path.join(parent,'surface_projection'), filename+'.raw.tif')

        # print(surface_proj.shape, surface_proj.dtype, surface_proj_file_path)
        Img(np.squeeze(raw_surface_proj), dimensions='hw').save(surface_proj_file_path)
        # ta_path = smart_name_parser(surface_proj_file_path,ordered_output=['TA'])

    ta_path = smart_name_parser(surface_proj_file_path, ordered_output='TA')

    height_map = result[2]

    # TODO compute heigh map quality (if out of frame --> need try yet another surface projection algorithm)
    # lui faire computer et sauver ce fichier

    if Z_FRAMES_TO_ADD is not None:
        height_map = height_map - Z_FRAMES_TO_ADD  # correct height map for added black frames --> then ideally I need also compute the heightmap quality that is
    Img(np.squeeze(height_map), dimensions='hw').save(os.path.join(ta_path, 'height_map.tif'))

    # rewind the stuff and do proj
    height_map_quality_test = surface_projection_from_height_map(Img(input_file), np.squeeze(height_map),
                                                                 channel=channel)  # TODO would need hack this to support the channel to be read for all images in case they are multichannel
    Img(height_map_quality_test, dimensions='hw').save(os.path.join(ta_path, 'height_map_quality_test.tif'))

    # save parent image properties in the db in the properties table so that they can be easily reused by TA
    update_db_properties_using_image_properties(input_file, ta_path)

    return surface_proj_file_path  # this is the file that needs be loaded in the other

# in fact denoise will not be most often on a proj image but will rather be
def denoise(deepTA, input_file):
    if isinstance(input_file, list):
        for file in input_file:
            try:
                denoise(deepTA, file)
            except:
                traceback.print_exc()
                print('Error: could not process file: "' + str(file) + '"')
        return
        # return
    parent, filename = smart_name_parser(
        input_file,
        ordered_output=['parent', 'short_no_ext'])
    # apparently it's a terrible idea to add frames and it makes sense btw, even though some models do lose data --> exclude those models!!!
    Z_FRAMES_TO_ADD = None
    result = predict_single_image(deepTA, input_file, Z_FRAMES_TO_ADD=Z_FRAMES_TO_ADD,
                                  input_normalization={'method': Img.normalization_methods[7], 'range': [2, 99.8],
                                                       'individual_channels': True, 'clip': False})
    denoised = result[0]
    Img(denoised, dimensions='hw').save(
        os.path.join(os.path.join(parent, 'denoised'), filename + '.tif'))  # store in pyta folder --> TODO


def predict_single_image(deepTA, input_file, TILE_WIDTH=256, TILE_HEIGHT=256, TILE_OVERLAP=32,
                         input_channel_of_interest=None, input_normalization=None, Z_FRAMES_TO_ADD=None,
                         SIZE_FILTER=None, hq_predictions='mean'):
    input_shape = deepTA.get_inputs_shape()
    output_shape = deepTA.get_outputs_shape()

    predict_parameters = {}

    predict_parameters["input_channel_of_interest"] = input_channel_of_interest
    predict_parameters["default_input_tile_width"] = TILE_WIDTH
    predict_parameters["default_input_tile_height"] = TILE_HEIGHT
    predict_parameters["default_output_tile_width"] = TILE_WIDTH
    predict_parameters["default_output_tile_height"] = TILE_HEIGHT
    predict_parameters["tile_width_overlap"] = TILE_OVERLAP
    predict_parameters["tile_height_overlap"] = TILE_OVERLAP
    # predict_parameters["hq_pred_options"] = "Use all augs (pixel preserving + deteriorating) (Recommended for segmentation)"  # not yet supported on current versions of epyseg
    predict_parameters[
        "hq_pred_options"] = 'Only use pixel preserving augs (Recommended for CARE-like models/surface extraction)'  # not yet supported on current versions of epyseg #None #'Only use pixel preserving augs (Recommended for CARE-like models/surface extraction)'  # not yet supported on current versions of epyseg
    predict_parameters["post_process_algorithm"] = None  # 'Keep first channel only'# None
    predict_parameters["input_normalization"] = input_normalization
    predict_parameters["filter"] = SIZE_FILTER

    # print('Z_FRAMES_TO_ADD',Z_FRAMES_TO_ADD)
    # print(input_normalization)

    predict_generator = deepTA.get_predict_generator(
        inputs=[input_file], input_shape=input_shape,
        output_shape=output_shape,
        clip_by_frequency=None,
        z_frames_to_add=Z_FRAMES_TO_ADD,  # addition especially for this one
        **predict_parameters)  # before was 0.05 which is bad especially with restored images

    # run for all the datagen
    # deepTA.predict(predict_generator, output_shape, predict_output_folder=predict_output_folder, batch_size=1,**predict_parameters)

    # hq_predictions = None
    # run one by one
    final_predict_generator = predict_generator.predict_generator()
    results = deepTA.predict_single(final_predict_generator, output_shape, batch_size=1, hq_predictions=hq_predictions,
                                    **predict_parameters)

    # print(results)
    # print(len(results), results[0].shape)
    #
    # plt.imshow((np.squeeze(results[0])))
    # plt.show()

    # fairly easy now --> I can have everything

    # could make this a class and get it to save the desired image where needed
    return results


# there is a big bug --> not same output as with the default GUI
if __name__ == '__main__':

    if False:
        # input_file = '/E/Sample_images/sample_images_pyta/surface_projection/210219.lif_t000.tif'
        input_file = '/E/Sample_images/sample_images_pyta/210219.lif_t000.tif'
        ta_path = smart_name_parser(input_file, 'TA')
        update_db_properties_using_image_properties(input_file,ta_path) # should do the job
        # try:
        #     tmp = Img(input_file)
        #     if has_metadata(tmp):
        #         db_path = os.path.join(ta_path, 'pyta.db')
        #         db = TAsql(db_path)
        #         try:
        #             # these are all the variables of the db I may want to keep...
        #             voxel_size_x = None
        #             voxel_size_y = None
        #             voxel_size_z = None
        #             voxel_z_over_x_ratio = None
        #             time = None
        #             creation_time = None
        #
        #             if 'vx' in tmp.metadata:
        #                 voxel_size_x = tmp.metadata['vx']
        #             if 'vy' in tmp.metadata:
        #                 voxel_size_y = tmp.metadata['vy']
        #             if 'vz' in tmp.metadata:
        #                 voxel_size_z = tmp.metadata['vz']
        #             if 'AR' in tmp.metadata:
        #                 voxel_z_over_x_ratio = tmp.metadata['AR']
        #             if 'time' in tmp.metadata:
        #                 time = tmp.metadata['time']
        #             if 'creation_time' in tmp.metadata:
        #                 creation_time = tmp.metadata['creation_time']
        #
        #             neo_data = {'voxel_size_x': voxel_size_x,
        #                         'voxel_size_y': voxel_size_y,
        #                         'voxel_size_z': voxel_size_z,
        #                         'voxel_z_over_x_ratio': voxel_z_over_x_ratio,
        #                         'time': time,
        #                         'creation_time': creation_time}
        #
        #             if db.exists('properties'):
        #                 # db exists --> update it rather than recreating everything
        #                 header, cols = db.run_SQL_command_and_get_results('SELECT * FROM properties', return_header=True)
        #                 cols = cols[0]
        #                 data = _to_dict(header, cols)
        #                 for key in list(neo_data.keys()):
        #                     if neo_data[key] is None:
        #                         if not key in data:
        #                             data[key] = neo_data[key]
        #                     else:
        #                         data[str(key)] = neo_data[key]
        #             else:
        #                 data = neo_data
        #
        #             data = {k: [v] for k, v in data.items()}
        #             db.create_and_append_table('properties', data)
        #         except:
        #             traceback.print_exc()
        #             print('An error occurred while reading properties from the image or writing them to the database')
        #         finally:
        #             try:
        #                 db.close()
        #             except:
        #                 pass
        #     # else:
        #     #     # image has no metadata --> nothing TODO --> I can just skip things
        #     #     pass
        #     del tmp
        # except:
        #     traceback.print_exc()
        #     print('Error could not save image properties to the TA database')

        import sys
        sys.exit(0)

    # faire un fast mode
    # could offer both offers

    # TODO also store the converted height map to smthg I

    use_cpu = True
    deepTA = EZDeepLearning(use_cpu=use_cpu)
    # best on my data
    # deepTA.load_or_build(model='/E/models/my_model/bckup/CARE_model_trained_MAE_in_epyseg_29-1_50steps_per_epoch_not_outstanding_though/surface_projection_model.h5')
    # deepTA.load_or_build(model='/E/models/my_model/bckup/CARE_model_trained_MSE_error_instead_MAE_in_epyseg_29-1/surface_projection_model.h5')
    # best consensus choice I guess # but I would most likely so far need to offer more choices
    # deepTA.load_or_build(model='/E/models/my_model/bckup/CARESEG_first_successful_retrain_on_CARE_data_very_very_good_surface_proj_but_not_for_all/surface_projection_model.h5') # maybe the best but really needs no addition of extra frames !!!!!# for this model extraframes are required otherwise signal is missing # very good on CARE too --> IN FACT QUITE GOOD ON BOTH --> MAYBE THE BEST BY DEFAULT STUFF # also has noise for the iamges of manue --> in that case there would be better choices
    # deepTA.load_or_build(model='/E/Sample_images/sample_images_pyta/test_merged_model.h5')  # TODO --> replace this by online model --> TODO


    # surface_proj_model = '/E/models/my_model/bckup/CARESEG_first_successful_retrain_on_CARE_data_very_very_good_surface_proj_but_not_for_all/surface_projection_model.h5'
    # denoiser_model = '/E/models/my_model/bckup/CARESEG_masked_data_and_masked_loss_to_prevent_identity_v1_210107_good/2D_denoiser_model.h5'  # not great on CARE great on others --> really need different denoisers depending on DATA --> ok --> though I doubt people will use CARE like data as input...
    # denoiser_model = '/E/models/my_model/bckup/CARESEG_another_normal_training_201216_colab_but_gave_crap/2D_denoiser_model.h5'# great for CARE not great for the others
    # denoiser_model = '/E/models/my_model/bckup/CARESEG_retrained_normally_201216/2D_denoiser_model.h5'# great for CARE not great for the others
    # denoiser_model = '/E/models/my_model/bckup/CARESEG_trained_on_CARE_data_with_dilation_good/2D_denoiser_model.h5'# great for CARE not great for the others
    surface_proj_model = 'SURFACE_PROJECTION_2'
    # surface_proj_model = 'SURFACE_PROJECTION_3'
    # surface_proj_model = 'SURFACE_PROJECTION_4'
    # denoiser_model = '2D_DENOISER'
    denoiser_model = '2D_DENOISEG'

    # pb is that
    model = create_surface_projection_denoise_and_height_map_combinatorial_model(surface_proj_model, denoiser_model)

    # shall I create a submodel ????

    print(model.summary(line_length=250))
    deepTA.model = model

    # deepTA.load_or_build(model='/E/models/my_model/bckup/FULL_CARE_MODEL_CONSENSUS_TRAINING_VERY_GOOD_140721/surface_projection_model.h5') # # nb adding frames to this one gives really crappy output --> do not do it!!! newly trained model --> really not bad in fact
    # deepTA.load_or_build(model='/E/models/my_model/bckup/CARESEG_second_correct_tested_210219_outstanding_denoise/surface_projection_model.h5')

    # best on CARE data
    # deepTA.load_or_build(model='/E/models/my_model/bckup/CARESEG_another_normal_training_201216_colab_but_gave_crap/surface_projection_model.h5')
    # deepTA.load_or_build(model='/E/models/my_model/bckup/CARESEG_masked_data_and_masked_loss_to_prevent_identity_v1_210107_good/surface_projection_model.h5') # works a bit on my data but less well than the one above
    # deepTA.load_or_build(model='/E/models/my_model/bckup/CARESEG_my_loss_mae_first_chan_dice_other_chans/surface_projection_model.h5')
    # deepTA.load_or_build(model='/E/models/my_model/bckup/CARESEG_trained_on_CARE_data_with_dilation_good/surface_projection_model.h5')
    # deepTA.load_or_build(model='/E/models/my_model/bckup/CARESEG_Zroll_models_trained_on_colab_201215/surface_projection_model.h5')

    # try find best denoiser!!!
    # not great # deepTA.load_or_build(model='/E/models/my_model/bckup/CARESEG_masked_data_and_masked_loss_to_prevent_identity_v1_210107_good/2D_denoiser_model.h5') # ok denoiser but not great on CARE!!!
    # deepTA.load_or_build(model='/E/models/my_model/bckup/CARESEG_another_normal_training_201216_colab_but_gave_crap/2D_denoiser_model.h5') # excellent on CARE not great on my own data
    # deepTA.load_or_build(model='/E/models/my_model/bckup/CARESEG_my_loss_mae_first_chan_dice_other_chans/2D_denoiser_model.h5') # not so bad on my data but weak signal
    # deepTA.load_or_build(model='/E/models/my_model/bckup/CARESEG_retrained_normally_201216/2D_denoiser_model.h5')
    # deepTA.load_or_build(model='/E/models/my_model/bckup/CARESEG_shuffle_n_roll_models_trained_on_colab_201215/2D_denoiser_model.h5') # a bit better on my data but not great (very weak in fact)
    # deepTA.load_or_build(model='/E/models/my_model/bckup/CARESEG_test_combined_loss_properly_made_with_reduce_mean_220105/2D_denoiser_model.h5')# second best on my data, very weak

    # second tests denoiser of my data
    # deepTA.load_or_build(model='/E/models/my_model/bckup/CARE_model_trained_MAE_in_epyseg_29-1_50steps_per_epoch_not_outstanding_though/2D_denoiser_model.h5') # good on my data but check on CARE ??? --> not great on CARE --> need really check individual ones --> TODO
    # deepTA.load_or_build(model='/E/models/my_model/bckup/CARESEG_first_correct_test/2D_denoiser_model.h5') # a bit artificial
    # deepTA.load_or_build(model='/E/models/my_model/bckup/CARESEG_first_successful_retrain_on_CARE_data_very_very_good_surface_proj_but_not_for_all/2D_denoiser_model.h5') # CRAP
    # deepTA.load_or_build(model='/E/models/my_model/bckup/CARESEG_second_correct_tested_210219_outstanding_denoise/2D_denoiser_model.h5')

    # common
    # deepTA.load_or_build(model='/E/models/my_model/bckup/CARESEG_shuffle_n_roll_models_trained_on_colab_201215/2D_denoiser_model.h5')
    # deepTA.load_or_build(model='/E/models/my_model/bckup/CARESEG_test_combined_loss_properly_made_with_reduce_mean_220105/2D_denoiser_model.h5')
    # deepTA.load_or_build(model='/E/models/my_model/bckup/CARESEG_trained_different_losses_for_channels_mae_bce/2D_denoiser_model.h5') # not too bad too --> maybe the most natural one ...
    # deepTA.load_or_build(model='/E/models/my_model/bckup/CARESEG_trained_on_CARE_data_with_dilation_good/2D_denoiser_model.h5') # BEST denoiser maybe best but not great --> still maybe best
    # deepTA.load_or_build(model='/E/models/my_model/bckup/CARESEG_Zroll_models_trained_on_colab_201215/2D_denoiser_model.h5') # maybe best # second best but still not so great, in a way specific denoisers need be used

    # if I just save model weights, it's gonna really take no space... --> maybe it's a good idea !!!
    # test of a consensus denoiser
    # deepTA.load_or_build(model='/home/aigouy/mon_prog/Python/epyseg_pkg/personal/training_deep/consensus_denoise_norm_input_output_none_training_150_epochs_lr_0.001_test2/2D_denoiser_model-0.h5') # not that bad in fact, check on care # good on my data but not great on CARE --> maybe offer different stuff
    # deepTA.load_or_build(model='/home/aigouy/mon_prog/Python/epyseg_pkg/personal/training_deep/consensus_denoise_norm_input_output_none_training_150_epochs_lr_0.001/2D_denoiser_model-0.h5') # not that bad in fact, check on care # good on my data but not great on CARE --> maybe offer different stuff
    # deepTA.load_or_build(model='/home/aigouy/mon_prog/Python/epyseg_pkg/personal/training_deep/2D_denoiser_model-0.h5') # not that bad in fact, check on care # good on my data but not great on CARE --> maybe offer different stuff
    # deepTA.load_or_build(model='/home/aigouy/mon_prog/Python/epyseg_pkg/personal/training_deep/2D_denoiser_model-0-0.h5') # not that bad in fact, check on care # good on my data but not great on CARE --> maybe offer different stuff
    # deepTA.load_or_build(model='/home/aigouy/mon_prog/Python/epyseg_pkg/personal/training_deep/2D_denoiser_model-0-0-0.h5') # not that bad in fact, check on care # good on my data but not great on CARE --> maybe offer different stuff
    # deepTA.load_or_build(model='/home/aigouy/mon_prog/Python/epyseg_pkg/personal/training_deep/consensus_denoise_normalized_input_output_training_100_epochs_lr_0.004_test5_MAE/2D_denoiser_model-0-0-0.h5') # VERY GOOD DENOISER but really ugly on CARE --> I will not manage --> do the ensemble code

    # test code find bug
    # deepTA.load_or_build(model='/home/aigouy/mon_prog/Python/epyseg_pkg/personal/training_deep/consensus_denoise_normalized_input_output_training_100_epochs_lr_0.004_test5_MAE/2D_denoiser_model-0-0-0.h5')  # VERY GOOD DENOISER but really ugly on CARE --> I will not manage --> do the ensemble code
    # the denoiser sucks
    # deepTA.load_or_build(model='/E/models/my_model/bckup/CARESEG_first_successful_retrain_on_CARE_data_very_very_good_surface_proj_but_not_for_all/CARESEG_model-0-0.h5')  # for this model extraframes are required otherwise signal is missing # very good on CARE too --> IN FACT QUITE GOOD ON BOTH --> MAYBE THE BEST BY DEFAULT STUFF # also has noise for the iamges of manue --> in that case there would be better choices

    # do for one image then ask which models to take --> see how much can be downloaded at the same time ???
    # doable with gitlab but will take a bit of space but ok maybe
    # maybe do a selection of the models...
    # faire une selection image par image --> pas trop bete en fait
    # faire max proj

    # TOD try my consensus denoiser just in case!!!

    # try all models on one image

    # others
    # deepTA.load_or_build(model='/E/models/my_model/bckup/CARE_masked_data_and_partial_masked_loss_to_limit_identity_but_not_remove_it_v1_210120_not_outstanding_CARESEG_is_better/surface_projection_model.h5')
    # deepTA.load_or_build(model='/E/models/my_model/bckup/CARESEG_trained_on_CARE_data_with_dilation_good/surface_projection_model.h5') # does not work well on my own data but does great on CARE

    # now really need to find a denoiser that I can append to then all done and since extra frames need do a subtraction to the file containing

    # TODO make it predict just for a single image and control everything --> in a way that would be much simpler to handle à la TA

    deepTA.get_loaded_model_params()
    deepTA.summary()

    IS_TA_OUTPUT_MODE = True  # stores as handCorrection.tif in the folder with the same name as the parent file without ext
    input_channel_of_interest = None  # 1  # assumes image is single channel or multichannel nut channel of interest is ch0, needs be changed otherwise, e.g. 1 for channel 1
    TILE_WIDTH = 512  # 256  # 128 # 64
    TILE_HEIGHT = 512  # 256  # 128 # 64
    TILE_OVERLAP = 32

    Z_FRAMES_TO_ADD = 5  # 5 #None
    # EPYSEG_PRETRAINING = 'Linknet-vgg16-sigmoid-v2'  # or 'Linknet-vgg16-sigmoid' for v1
    SIZE_FILTER = None  # 100 # set to 100 to get rid of cells having pixel area < 100 pixels

    # input_normalization = {'method': 'Rescaling (min-max normalization)', 'range': [0, 1], 'individual_channels': True}
    # input_normalization = {'method': Img.normalization_methods[7], 'range': [2, 99.8],'individual_channels': True, 'clip': False}
    input_normalization = None

    # INPUT_FOLDER = '/path/to/files_to_segment/'

    # surface proj tests

    # INPUT_FILE = '/E/Sample_images/sample_images_pyta/exposures_1_P04.tif'
    # INPUT_FILE = '/E/Sample_images/sample_images_pyta/210219.lif_t000.tif'
    # INPUT_FILE = '/E/Sample_images/sample_images_pyta/Image49.lsm'
    # INPUT_FILE = '/E/Sample_images/sample_images_pyta/Image4.lsm'

    # INPUT_FILE = ['/E/Sample_images/sample_images_pyta/exposures_1_P04.tif', '/E/Sample_images/sample_images_pyta/210219.lif_t000.tif', '/E/Sample_images/sample_images_pyta/Image49.lsm', '/E/Sample_images/sample_images_pyta/Image4.lsm']

    INPUT_FILE = ['/E/Sample_images/sample_images_denoise_manue/210922_armGFP_suz_40-54hAPF_ON.lif - Series004_t002_reveresed.tif',
                    # '/E/Sample_images/sample_images_pyta/exposures_1_P04.tif',
                  # '/E/Sample_images/sample_images_pyta/210219.lif_t000.tif',
                  # '/E/Sample_images/sample_images_pyta/Image48.lsm',
                  #   '/E/Sample_images/sample_images_pyta/Image4.lsm'
                  ]

    # denoiser tests
    # INPUT_FILE = '/E/Sample_images/sample_images_pyta/focused_Series012.png'
    # INPUT_FILE = '/E/Sample_images/sample_images_pyta/Image29.tif'
    # INPUT_FILE = '/E/Sample_images/sample_images_pyta/210219.lif_t000/handCorrection.tif'
    # INPUT_FILE = '/E/Sample_images/sample_images_pyta/Image49/handCorrection.tif'
    # INPUT_FILE = '/E/Sample_images/sample_images_pyta/exposures_1_P04/handCorrection.tif'

    # result = predict_single_image(deepTA, INPUT_FILE, TILE_WIDTH=TILE_WIDTH, TILE_HEIGHT=TILE_HEIGHT, TILE_OVERLAP=TILE_OVERLAP, input_normalization=input_normalization, SIZE_FILTER=SIZE_FILTER, Z_FRAMES_TO_ADD=Z_FRAMES_TO_ADD, input_channel_of_interest=input_channel_of_interest)
    # print(result[0].shape)

    # TODO create the library the proper way also see how I can connect directly the denoiser
    # surface_projection_pyta(deepTA, INPUT_FILE,recursion_for_denoising=4)
    surface_projection_pyta(deepTA, INPUT_FILE, recursion_for_denoising=1)

    # finalize all now then ok, transfer files to the stuff
    # sinon dire ou sauver le fichier et tt faire comme il faut
    # sinon si max proj --> sauver dans le fichier correspondant --> non en fait c'est ok!
